from math import isclose

import pytest

from tidyms2.chem import PTABLE, formula
from tidyms2.chem.exceptions import IsotopeError


@pytest.mark.parametrize(
    "formula_str,p_open,p_close",
    [
        ("[Cr[H2O]6]3+", 0, 9),
        ("[C9H11NO2]", 0, 9),
        ("C9H11N(17O)2", 6, 10),
        ("[Cr[(2H)2O]6]3+", 3, 10),
    ],
)
def test_find_matching_parenthesis_valid_input(formula_str, p_open, p_close):
    test_p_close = formula._find_matching_parenthesis(formula_str, p_open)
    assert test_p_close == p_close


@pytest.mark.parametrize(
    "formula_str,formula_without_charge,charge",
    [
        ("H2O", "H2O", 0),
        ("(13C)", "(13C)", 0),
        ("[CO3]2-", "CO3", -2),
        ("[Cr[H2O]6]3+", "Cr[H2O]6", 3),
        ("[C9H11NO2]", "[C9H11NO2]", 0),
        ("CO-", "CO", -1),
        ("[H2O]+", "[H2O]", 1),
        ("H2O+", "H2O", 1),
    ],
)
def test_parse_charge_valid_input(formula_str, formula_without_charge, charge):
    test_formula_without_charge, q = formula._parse_charge(formula_str)
    assert test_formula_without_charge == formula_without_charge
    assert charge == q


@pytest.mark.parametrize("formula_str", ["SO42-"])
def test_parse_charge_invalid_input(formula_str):
    with pytest.raises(formula.InvalidFormula):
        _, q = formula._parse_charge(formula_str)


@pytest.mark.parametrize(
    "formula_str,ind,token_type",
    [
        ("H2O", 0, 0),
        ("H2(34S)O4", 2, 1),
        ("[Cr(H2O)6]3+", 3, 2),
        ("[Fe[CN]6]4-", 3, 2),
    ],
)
def test_get_token_type(formula_str, ind, token_type):
    test_token_type = formula._get_token_type(formula_str, ind)
    assert test_token_type == token_type


@pytest.mark.parametrize(
    "formula_str,ind,coeff,new_ind",
    [
        ("H2O", 3, 1, 3),
        ("CO2", 1, 1, 1),
        ("C9H11NO2", 3, 11, 5),
    ],
)
def test_get_coefficient_valid_input(formula_str, ind, coeff, new_ind):
    test_coeff, test_ind = formula._get_coefficient(formula_str, ind)
    assert coeff == test_coeff
    assert test_ind == new_ind


@pytest.mark.parametrize(
    "formula_str,ind,new_ind,element",
    [
        ("H2O", 0, 2, "H"),
        ("H2O", 2, 3, "O"),
        ("C9H11NO2", 5, 6, "N"),
        ("C9H11N(17O)2", 5, 6, "N"),
        ("Cr(H2O)6", 0, 2, "Cr"),
    ],
)
def test_tokenize_element_valid_input(formula_str, ind, new_ind, element):
    token, test_index = formula._tokenize_element(formula_str, ind, PTABLE)
    assert test_index == new_ind
    isotope = PTABLE.get_element(element).monoisotope
    assert isotope in token


@pytest.mark.parametrize(
    "formula_str,ind,isotope_str,new_ind",
    [
        ("(13C)O2", 0, "13C", 5),
        ("C9H11(15N)2O2", 5, "15N", 11),
        ("C6H12O5(18O)", 7, "18O", 12),
        ("C6H12O4(18O)2", 7, "18O", 13),
    ],
)
def test_tokenize_isotope_valid_input(formula_str, ind, isotope_str, new_ind):
    token, test_index = formula._tokenize_isotope(formula_str, ind, PTABLE)
    isotope = PTABLE.get_isotope(isotope_str)
    assert test_index == new_ind
    assert isotope in token


@pytest.mark.parametrize(
    "f_str,composition",
    [
        ("H2O", {"1H": 2, "16O": 1}),
        ("(13C)O2", {"13C": 1, "16O": 2}),
        ("C9H11(15N)2O2", {"12C": 9, "1H": 11, "15N": 2, "16O": 2}),
        ("C9H11N2O2", {"12C": 9, "1H": 11, "14N": 2, "16O": 2}),
        ("Cr[(2H)2O]6", {"52Cr": 1, "2H": 12, "16O": 6}),
    ],
)
def test_tokenize_formula(f_str, composition):
    composition = {PTABLE.get_isotope(k): v for k, v in composition.items()}
    test_composition = formula._parse_formula(f_str, PTABLE)
    for isotope in composition:
        assert composition[isotope] == test_composition[isotope]


def test_arg_sort_elements():
    symbols = ["Cd", "C", "H", "H", "O", "O", "S", "B"]
    a = [60, 12, 2, 1, 16, 17, 32, 7]
    sorted_ind = [7, 1, 0, 3, 2, 4, 5, 6]
    assert sorted_ind == formula._arg_sort_elements(symbols, a)


@pytest.mark.parametrize("charge,charge_str", [(1, "+"), (2, "2+"), (-1, "-"), (-4, "4-")])
def test_get_charge_str(charge, charge_str):
    test_charge_str = formula._get_charge_str(charge)
    assert test_charge_str == charge_str


@pytest.mark.parametrize(
    "f,f_str",
    [
        (formula.Formula("CO2"), "CO2"),
        (formula.Formula("(13C)C2H6O3"), "C2(13C)H6O3"),
        (formula.Formula("C24H46SPN(18O)2"), "C24H46N(18O)2PS"),
        (formula.Formula("[Cr(H2O)6]3+"), "[H12CrO6]3+"),
        (formula.Formula("CH3CH2CH3"), "C3H8"),
        (formula.Formula("F2"), "F2"),
    ],
)
def test_get_formula_str(f, f_str):
    test_f_str = str(f)
    assert test_f_str == f_str


@pytest.mark.parametrize("f_str", ["(CO2", "#H2O"])
def test_parse_formula_invalid_formula(f_str):
    with pytest.raises(formula.InvalidFormula):
        formula.Formula(f_str)


@pytest.mark.parametrize("f_str", ["(14C)O2", "(3H)2O"])
def test_parse_formula_invalid_isotope(f_str):
    with pytest.raises(IsotopeError):
        formula.Formula(f_str)


@pytest.fixture
def formula_data():
    formula_str = ["CO2", "H2O", "F2"]
    nominal = [44, 18, 38]
    exact = [43.9898, 18.0106, 37.9968]
    return formula_str, nominal, exact


def test_get_exact_mass(formula_data):
    formula_str, _, exact = formula_data
    for f_str, e in zip(formula_str, exact):
        assert abs(formula.Formula(f_str).get_exact_mass() - e) < 0.0001


def test_get_nominal_mass(formula_data):
    formula_str, nominal, _ = formula_data
    for f_str, n in zip(formula_str, nominal):
        assert formula.Formula(f_str).get_nominal_mass() == n


def test_get_isotopic_envelope_n_1_returns_abundance_1():
    f = formula.Formula("C24H46SPN(18O)2")
    envelope = f.get_isotopic_envelope(n=1)
    assert len(envelope.p) == 1
    assert isclose(envelope.p[0], 1.0)


def test_formula_from_dictionary():
    composition = {"C": 1, "17O": 2, "H": 2}
    charge = 1
    f = formula.Formula(composition, charge)
    for k in composition:
        assert PTABLE.get_isotope(k) in f.composition
    assert charge == f.charge


def test_formula_from_dictionary_invalid_isotope():
    composition = {"C": 1, "G": 4}
    charge = 1
    with pytest.raises(IsotopeError):
        formula.Formula(composition, charge)


def test_formula_from_dictionary_invalid_isotope_type():
    composition = {4: 1, "G": 4}
    charge = 1
    with pytest.raises(ValueError):
        formula.Formula(composition, charge)  # type: ignore


@pytest.mark.parametrize(
    "composition,q",
    [
        [{"C": -1, "H": 4}, 1],
        [{"C": 1, "H": 4}, 0.5],
    ],
)
def test_formula_from_dictionary_invalid_coefficient(composition, q):
    with pytest.raises(ValueError):
        formula.Formula(composition, q)


def test_Formula_add():
    f1 = formula.Formula("H2O")
    f2 = formula.Formula("CO2")
    f_sum = f1 + f2
    expected = formula.Formula("H2CO3")
    assert expected == f_sum


def test_Formula_subtract_valid():
    f1 = formula.Formula("C6H12O6")
    f2 = formula.Formula("CO2")
    f_diff = f1 - f2
    expected = formula.Formula("C5H12O4")
    assert expected == f_diff


def test_Formula_subtract_valid_zero_coeff():
    f1 = formula.Formula("C4H8O2")
    f2 = formula.Formula("CO2")
    f_diff = f1 - f2
    expected = formula.Formula("C3H8")
    assert expected == f_diff


def test_Formula_subtract_invalid_coeff():
    f1 = formula.Formula("C4H8O")
    f2 = formula.Formula("CO2")
    with pytest.raises(ValueError):
        f1 - f2  # type: ignore



"""Helpers classes and functions for unit tests."""

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from random import randint, random, uniform

import numpy
from pydantic import Field, computed_field
from typing_extensions import Self

from tidyms2.chem import Formula
from tidyms2.core import models
from tidyms2.core.enums import MSInstrument, Polarity, SeparationMode
from tidyms2.core.matrix import DataMatrix
from tidyms2.core.models import AnnotableFeature, FeatureGroup, GroupAnnotation, IsotopicEnvelope, Roi, Sample
from tidyms2.core.operators.assay import AnnotationPatcher, DescriptorPatcher
from tidyms2.core.operators.sample import FeatureExtractor, FeatureTransformer, RoiExtractor, RoiTransformer
from tidyms2.core.registry import operator_registry
from tidyms2.core.storage import AssayStorage


class ConcreteRoi(Roi):
    data: list[float] = list()

    def equal(self, other: ConcreteRoi) -> bool:
        """Compare equal."""
        return self.data == other.data


class ConcreteFeature(AnnotableFeature[ConcreteRoi]):
    data_mz: float = Field(repr=False)
    """A proxy variable for the feature m/z."""

    data_area: float = Field(repr=False, default=100.0)
    """A proxy variable for the feature area"""

    @computed_field(repr=False)
    @cached_property
    def custom_descriptor(self) -> float:
        return 100.0

    @computed_field(repr=False)
    @cached_property
    def area(self) -> float:
        return self.data_area

    @computed_field(repr=False)
    @cached_property
    def mz(self) -> float:
        return self.data_mz

    @computed_field(repr=False)
    @cached_property
    def height(self) -> float:
        return self.data_area

    def equal(self, other: ConcreteFeature) -> bool:
        return self.data_mz == other.data_mz

    def compare(self, other: Self) -> float:
        return 1.0

    @staticmethod
    def compute_isotopic_envelope(*features) -> IsotopicEnvelope:
        total_area = sum([x.area for x in features])
        mz = [x.mz for x in features]
        p = [x.area / total_area for x in features]
        return IsotopicEnvelope(mz=mz, p=p)


@operator_registry.register()
class DummyRoiExtractor(RoiExtractor[ConcreteRoi, ConcreteFeature]):
    n_roi: int = 3
    """The number of dummy ROI to extract."""

    param2: str = "default"

    def extract_rois(self, sample: Sample) -> list[ConcreteRoi]:
        return [create_roi(sample) for _ in range(self.n_roi)]

    def pre_apply(self):
        """Test pre apply functionality."""
        pass

    def post_apply(self):
        """Test post apply functionality."""
        pass

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()


@operator_registry.register()
class DummyRoiTransformer(RoiTransformer[ConcreteRoi, ConcreteFeature]):
    max_length: int = 2
    """Crop ROI length to max length. ROIs with length greater than this value are deleted."""

    param2: str = "default"

    def transform_roi(self, roi: ConcreteRoi) -> ConcreteRoi | None:
        if len(roi.data) <= self.max_length:
            roi.data = roi.data[: self.max_length]
            return roi
        return None

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()


@operator_registry.register()
class DummyFeatureExtractor(FeatureExtractor[ConcreteRoi, ConcreteFeature]):
    n_features: int = 2
    """The number of features to extract from each ROI."""

    param2: str = "default"

    def extract_features(self, roi: ConcreteRoi) -> list[ConcreteFeature]:
        return [create_feature(roi) for _ in range(self.n_features)]

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()


@operator_registry.register()
class DummyFeatureTransformer(FeatureTransformer[ConcreteRoi, ConcreteFeature]):
    feature_value: int = 5
    """The value to set in feature data."""

    param2: str = "default"

    def transform_feature(self, feature: ConcreteFeature) -> ConcreteFeature | None:
        if feature.data_mz <= self.feature_value:
            feature.data_mz = self.feature_value
            return feature
        return None

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()


@operator_registry.register()
class DummyAnnotationPatcher(AnnotationPatcher):
    """Set annotation group to a fixed values in all features."""

    group: int = 0

    def compute_patches(self, data: AssayStorage) -> list[models.AnnotationPatch]:
        annotations = data.fetch_annotations()
        return [models.AnnotationPatch(id=x.id, field="group", value=0) for x in annotations]

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    def pre_apply(self):
        """Test pre apply functionality."""
        pass

    def post_apply(self):
        """Test post apply functionality."""
        pass


@operator_registry.register()
class DummyDescriptorPatcher(DescriptorPatcher):
    """Set descriptor to a fixed value in all features."""

    patch: float = 500.0
    descriptor: str = "custom_descriptor"

    def compute_patches(self, data: AssayStorage) -> list[models.DescriptorPatch]:
        annotations = data.fetch_annotations()
        return [models.DescriptorPatch(id=x.id, descriptor=self.descriptor, value=self.patch) for x in annotations]

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity) -> Self:
        return cls()

    def pre_apply(self):
        """Test pre apply functionality."""
        pass

    def post_apply(self):
        """Test post apply functionality."""
        pass


def create_sample(
    path: Path | None = None,
    suffix: int = 0,
    group: str = "",
    order: int | None = None,
    batch: int = 0,
    create_file: bool = True,
) -> models.Sample:
    """Create a sample model.

    :param path: path to the sample. If ``None`` is used, then the current working directory is used.
    :param suffix: a suffix added to the sample name. Samples are named using `sample-{suffix}`.
    :param group: sample group passed to sample metadata.
    :param order: sample run order passed to sample metadata. If not defined, the `suffix` parameter
        is used.
    :param batch: the analytical batch passed to sample metadata.
    :param create_file: if set to ``True``, touch the file path.

    """
    file = (path or Path.cwd()) / f"sample-{suffix}.mzML"
    if create_file:
        file.touch()
    sample = models.Sample(path=file, id=file.stem)
    sample.meta.group = group
    sample.meta.order = order or suffix
    sample.meta.batch = batch
    return sample


def create_roi(sample: Sample) -> ConcreteRoi:
    data = [random() for _ in range(5)]
    return ConcreteRoi(data=data, sample=sample)


def create_feature(roi: ConcreteRoi) -> ConcreteFeature:
    data = randint(0, 10)
    return ConcreteFeature(roi=roi, data_mz=data)


def create_features_from_formula(formula_str: str, sample: Sample, n_isotopologues: int = 10) -> list[ConcreteFeature]:
    formula = Formula(formula_str)
    env = formula.get_isotopic_envelope(n_isotopologues)
    q = abs(formula.charge) if formula.charge else 1
    features = list()
    for Mk, pk in zip(env.mz, env.p):
        features.append(ConcreteFeature(data_mz=Mk / q, data_area=pk, roi=ConcreteRoi(sample=sample)))
    return features


def create_feature_group(group: int) -> FeatureGroup:
    ann = GroupAnnotation(label=group)
    descriptors = {"mz": uniform(100.0, 1000.0)}
    return FeatureGroup(group=group, annotation=ann, descriptors=descriptors)


def create_data_matrix(n_samples: int, n_features: int, sample_path: Path) -> DataMatrix:
    samples = [create_sample(sample_path, k) for k in range(n_samples)]
    features = [create_feature_group(k) for k in range(n_features)]
    data = numpy.random.normal(loc=100.0, size=(n_samples, n_features))
    return DataMatrix(samples, features, data)



import pathlib

import numpy as np
import pytest
from sklearn.cluster import DBSCAN

from tidyms2.annotation import correspondence
from tidyms2.core.models import Annotation, AnnotationPatch, Sample
from tidyms2.core.utils.common import create_id


@pytest.mark.parametrize(
    "n_sample,n_cluster,max_size", [[20, 2, 10], [100, 4, 125], [200, 25, 1500], [200, 10, 20000]]
)
def test_cluster_db_scan_solution_is_equal_to_sklearn_solution(n_sample, n_cluster, max_size):
    X1 = np.arange(n_sample)
    X = np.vstack((X1, X1)).T
    X = np.repeat(X, n_cluster, axis=0)
    X = np.random.permutation(X)

    eps = 0.1
    min_samples = round(n_sample * 0.2)
    test_cluster = correspondence.cluster_dbscan(X, eps, min_samples, max_size)
    dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric="chebyshev")
    dbscan.fit(X)
    expected_cluster = dbscan.labels_
    assert np.array_equal(test_cluster, expected_cluster)


class TestEstimateNSpecies:
    @pytest.mark.parametrize(
        "min_samples,expected",
        [[1, np.array([2, 2])], [2, np.array([2, 2])], [3, np.array([0, 0])]],
    )
    def test_one_class(self, min_samples, expected):
        samples = np.array([0] * 4 + [1] * 4)  # 8 features detected in total in two samples
        clusters = np.array(([0] * 2 + [1] * 2) * 2)  # two clusters
        n_clusters = 2
        # two species in two clusters are expected
        res = correspondence._estimate_n_species_one_group(samples, clusters, min_samples, n_clusters)
        assert np.array_equal(res, expected)

    def test_estimate_n_species_multiple_groups(self):
        samples = np.array([0] * 4 + [1] * 4 + [2] * 4)  # 12 features in three samples
        clusters = np.array(([0] * 2 + [1] * 2) * 3)  # two clusters
        classes = np.array([0] * 8 + [1] * 4)  # two groups
        min_dr = 0.5
        # two species in two clusters are expected
        expected = {0: 2, 1: 2}
        include_classes = [0, 1]
        samples_per_class = {0: 2, 1: 1}

        res = correspondence._estimate_gmm_n_components(
            samples, clusters, classes, samples_per_class, include_classes, min_dr
        )
        assert res == expected


class TestEstimateDBSCANMinSample:
    @pytest.fixture
    def samples_per_class(self):
        res = {0: 8, 1: 16, 2: 24}
        return res

    def test_no_include_groups_ok(self, samples_per_class):
        min_fraction = 0.25
        include_classes = None
        test_min_samples = correspondence._estimate_dbscan_min_sample(samples_per_class, include_classes, min_fraction)
        expected_min_samples = round(sum(samples_per_class.values()) * min_fraction)
        assert expected_min_samples == test_min_samples

    def test_include_groups_ok(self, samples_per_class):
        min_fraction = 0.25
        include_classes = [0, 1]
        test_min_samples = correspondence._estimate_dbscan_min_sample(samples_per_class, include_classes, min_fraction)
        n_include = [v for k, v in samples_per_class.items() if k in include_classes]
        expected_min_samples = round(min(n_include) * min_fraction)
        assert expected_min_samples == test_min_samples


class TestProcessCluster:
    def test_one_species(self):
        np.random.seed(1234)
        # features
        n = 200
        X = np.random.normal(size=(n, 2))
        samples = np.arange(n)

        # add noise
        n_noise = 10
        noise = np.random.normal(size=(n_noise, 2), loc=4)
        X = np.vstack((X, noise))
        s_noise = np.random.choice(samples, size=n_noise)
        samples = np.hstack((samples, s_noise))

        expected = np.array([0] * n + [-1] * n_noise)

        n_species = 1
        max_deviation = 4
        labels, score = correspondence.split_cluster_gmm(X, samples, n_species, max_deviation)
        assert np.array_equal(labels, expected)

    def test_two_species(self):
        np.random.seed(1234)
        # features
        n = 200
        x_list = list()
        s_list = list()
        for loc in [0, 4]:
            x_list.append(np.random.normal(size=(n, 2), loc=loc))
            s_list.append(np.arange(n))

        # add noise
        n_noise = 10
        x_list.append(np.random.normal(size=(n_noise, 2), loc=8))
        X = np.vstack(x_list)
        s_list.append(np.random.choice(s_list[0], size=n_noise))
        samples = np.hstack(s_list)

        n_species = 2
        max_deviation = 4
        labels, score = correspondence.split_cluster_gmm(X, samples, n_species, max_deviation)

        # create expected array: it is created after actual computation because
        # it is not possible to know which label is assigned to each cluster
        expected_first_label, expected_second_label = (0, 1) if labels[0] == 0 else (1, 0)
        expected = np.array([expected_first_label] * n + [expected_second_label] * n + [-1] * n_noise)

        assert np.array_equal(labels, expected)


def test_match_features():
    np.random.seed(1234)
    descriptors = dict()
    n_noise = 10
    n_samples = 200
    samples = [Sample(id=f"sample-{x}", path=pathlib.Path(".")) for x in range(n_samples)]
    annotations = list()
    for n, loc in [(200, 0), (200, 4), (n_noise, 8)]:
        mz_list = descriptors.setdefault("mz", list())
        mz_list.extend(np.random.normal(size=n, loc=loc))

        rt_list = descriptors.setdefault("rt", list())
        rt_list.extend(np.random.normal(size=n, loc=loc))

        annotations.extend(Annotation(sample_id=f"sample-{x}", roi_id=create_id()) for x in range(n))

    tolerance = {"mz": 2.0, "rt": 2.0}

    params = correspondence.FeatureCorrespondenceParameters(max_deviation=4.0)
    actual = correspondence.match_features(descriptors, annotations, samples, tolerance, params)

    # expected result is created after actual computation because it is not
    # possible to know which label is assigned to each cluster before the algorithm runs
    expected_first_label, expected_second_label = (0, 1) if actual[0].value == 0 else (1, 0)
    expected_labels = np.array([expected_first_label] * 200 + [expected_second_label] * 200 + [-1] * n_noise)
    expected = [AnnotationPatch(id=x.id, field="group", value=y) for x, y in zip(annotations, expected_labels)]

    assert actual == expected


class TestComputeFeatureGroupCentroids:
    n_descriptors = 2

    def test_no_data_ok(self):
        X = np.array([], dtype=float)
        groups = np.array([], dtype=int)
        unique_groups, centroids = correspondence._compute_feature_group_centroids(X, groups)

        assert unique_groups.size == 0
        assert unique_groups.dtype == int

        assert centroids.size == 0
        assert centroids.dtype == float

    @pytest.mark.parametrize("n_ft", [1, 5, 10])
    def test_single_group_ok(self, n_ft):
        X = np.ones(shape=(n_ft, self.n_descriptors), dtype=float)
        groups = np.zeros(n_ft, dtype=int)

        expected_centroids = np.array([1.0 for _ in range(self.n_descriptors)])
        expected_unique_groups = np.array([0])
        actual_unique_groups, actual_centroids = correspondence._compute_feature_group_centroids(X, groups)

        assert np.array_equal(expected_unique_groups, actual_unique_groups)
        assert np.allclose(expected_centroids, actual_centroids)

    @pytest.mark.parametrize("n_ft1", [1, 5, 10])
    def test_multiple_group_ok(self, n_ft1):
        n_ft2 = 5

        ft1_constant = 1.0
        ft2_constant = 2.0
        X = np.vstack(
            (
                ft1_constant * np.ones(shape=(n_ft1, self.n_descriptors), dtype=float),
                ft2_constant * np.ones(shape=(n_ft2, self.n_descriptors), dtype=float),
            )
        )
        groups = np.hstack([np.zeros(n_ft1, dtype=int), np.ones(n_ft2, dtype=int)])

        expected_unique_groups = np.array([0, 1])
        expected_centroids = np.array([[x] * self.n_descriptors for x in [ft1_constant, ft2_constant]])

        actual_unique_groups, actual_centroids = correspondence._compute_feature_group_centroids(X, groups)

        assert np.array_equal(expected_unique_groups, actual_unique_groups)
        assert np.allclose(expected_centroids, actual_centroids)


class TestFindMergeCandidates:
    n_descriptors = 2
    tol = 0.1

    def test_no_candidates(self):
        centroids = np.array([[0.0, 0.0], [1.0, 1.0]])
        groups = np.array([0, 1])

        candidates = correspondence._find_merge_candidates(centroids, groups, self.tol)
        assert not candidates

    def test_one_candidate(self):
        centroids = np.array([[0.0, 0.0], [1.0, 1.0], [1.05, 1.05]])
        groups = np.array([0, 1, 2])

        actual = correspondence._find_merge_candidates(centroids, groups, self.tol)
        expected = [(1, 2)]
        assert actual == expected

    def test_two_candidates(self):
        centroids = np.array([[0.0, 0.0], [1.0, 1.0], [1.05, 1.05], [2.0, 2.0], [2.05, 2.05]])
        groups = np.array([0, 1, 2, 3, 4])

        actual = correspondence._find_merge_candidates(centroids, groups, self.tol)
        expected = [(1, 2), (3, 4)]
        assert actual == expected

    def test_two_consecutive_candidates(self):
        centroids = np.array([[0.0, 0.0], [1.0, 1.0], [1.05, 1.05], [1.075, 1.1]])
        groups = np.array([0, 1, 2, 3])

        actual = correspondence._find_merge_candidates(centroids, groups, self.tol)
        expected = [(1, 2), (2, 3)]
        assert actual == expected


class TestMergeIfClose:
    # samples with 0.5 overlap
    group1 = 1
    samples1 = {"s1", "s2", "s3", "s4"}
    group2 = 2
    samples2 = {"s1", "s2"}

    def test_no_merge_if_sample_overlap_is_higher_than_max(self):
        max_sample_overlap = 0.25

        expected_merger, expected_merged = -1, -1
        actual_merger, actual_merged = correspondence._merge_if_close(
            self.group1, self.samples1, self.group2, self.samples2, max_sample_overlap
        )

        assert expected_merger == actual_merger
        assert expected_merged == actual_merged

    def test_no_merge_symmetric_result(self):
        max_sample_overlap = 0.25
        actual_merger, actual_merged = correspondence._merge_if_close(
            self.group1, self.samples1, self.group2, self.samples2, max_sample_overlap
        )
        swapped_merger, swapped_merged = correspondence._merge_if_close(
            self.group2, self.samples2, self.group1, self.samples1, max_sample_overlap
        )
        assert actual_merger == swapped_merger
        assert actual_merged == swapped_merged

    def test_merge_if_sample_overlap_is_lower_than_max(self):
        max_sample_overlap = 0.6

        expected_merger, expected_merged = 1, 2

        actual_merger, actual_merged = correspondence._merge_if_close(
            self.group1, self.samples1, self.group2, self.samples2, max_sample_overlap
        )

        assert expected_merger == actual_merger
        assert expected_merged == actual_merged

    def test_merge_symmetric_result(self):
        max_sample_overlap = 0.6
        actual_merger, actual_merged = correspondence._merge_if_close(
            self.group1, self.samples1, self.group2, self.samples2, max_sample_overlap
        )
        swapped_merger, swapped_merged = correspondence._merge_if_close(
            self.group2, self.samples2, self.group1, self.samples1, max_sample_overlap
        )
        assert actual_merger == swapped_merger
        assert actual_merged == swapped_merged


def test_merge_candidates():
    n = 2
    ft_constants = [0.0, 1.0, 2.0, 2.05]
    n_descriptors = 2

    X = np.vstack([x * np.ones(shape=(n, n_descriptors)) for x in ft_constants])
    groups = np.array([0, 0, 1, 1, 2, 2, 3, 3])
    samples = ["s1", "s2", "s1", "s2", "s1", "s2", "s3", "s4"]

    max_sample_overlap = 0.5
    tol = 0.1

    expected = np.array([0, 0, 1, 1, 2, 2, 2, 2])
    actual = correspondence.merge_feature_groups(X, groups, samples, tol, max_sample_overlap)

    assert np.array_equal(actual, expected)



from math import isclose

import pytest
from pydantic import ValidationError

from tidyms2.simulation.base import AbundanceSpec, InstrumentResponseSpec, MeasurementNoiseSpec


class TestAbundanceSpec:
    def test_sample_abundance_default_settings_return_constant_value(self):
        spec = AbundanceSpec()
        assert isclose(spec.mean, spec.sample_abundance())

    def test_sample_abundance_with_std_return_different_values(self):
        spec = AbundanceSpec(std=5.0)

        # first collect multiple values and check that they are different
        values = list()
        n = 10
        for _ in range(n):
            values.append(spec.sample_abundance())
        assert len(set(values)) > 1

        tol = 5 * spec.std  # p ~ 0.99999 of all values being within this tolerance
        assert isclose(sum(values) / n, spec.mean, abs_tol=tol)

    def test_sample_abundance_with_prevalence_return_mean_or_zero(self):
        spec = AbundanceSpec(prevalence=0.5)

        values = [spec.sample_abundance() for _ in range(30)]
        assert all(isclose(spec.mean, x) or isclose(0.0, x) for x in values)
        # check at least one zero (p ~ 1e-10 of all values being non zero)
        assert sum(isclose(x, 0.0) for x in values) >= 1

    def test_sample_abundance_with_prevalence_and_std(self):
        spec = AbundanceSpec(prevalence=0.5)

        n = 30
        values = [spec.sample_abundance() for _ in range(n)]
        non_zeros = [x for x in values if not isclose(x, 0.0)]

        # check at least one non_zero (p ~ 1e-10 of all values being non zero)
        assert len(non_zeros) < n

        # then check that non zero values follow gaussian dist
        tol = 5 * spec.std  # p ~ 0.9997 of all values being within this tolerance
        non_zero_values_mean = sum(non_zeros) / len(non_zeros)
        assert isclose(non_zero_values_mean, spec.mean, abs_tol=tol)


class TestMeasurementNoiseSpec:
    def test_sample_noise_no_snr_returns_zero(self):
        spec = MeasurementNoiseSpec()
        assert isclose(0.0, spec.sample_noise(100.0, 1.0))

    def test_snr_lower_than_min_snr_raises_error(self):
        with pytest.raises(ValidationError):
            MeasurementNoiseSpec(base_snr=5.0, min_snr=10.0)

    def test_compute_snr_no_base_snr_return_none(self):
        spec = MeasurementNoiseSpec(base_snr=None)
        assert spec.compute_snr(1.0) is None

    def test_compute_snr_greater_than_min_snr(self):
        spec = MeasurementNoiseSpec(base_snr=100.0, min_snr=10.0)
        pk = 0.20
        snr = spec.compute_snr(pk)
        assert snr is not None
        assert snr > spec.min_snr

    def test_compute_snr_lower_than_min_snr_returns_min_snr(self):
        spec = MeasurementNoiseSpec(base_snr=100.0, min_snr=10.0)
        pk = 0.01
        snr = spec.compute_snr(pk)
        assert snr is not None
        assert snr == spec.min_snr

    def test_sample_noise_default_configuration_returns_zero(self):
        spec = MeasurementNoiseSpec()
        for _ in range(10):
            assert isclose(0.0, spec.sample_noise(1000.0, 1.0))

    def test_sample_noise_is_within_expected_values(self):
        snr = 100.0
        spec = MeasurementNoiseSpec(base_snr=snr)
        signal = 1000.0
        pk = 1.0

        tol = 5 * signal / snr  # p ~ 0.999999 of noise being within this value

        noise = spec.sample_noise(signal, pk)
        assert isclose(noise, 0.0, abs_tol=tol)


class TestInstrumentResponseSpec:
    def test_get_interbatch_with_default_configuration_return_one(self):
        spec = InstrumentResponseSpec()
        for k in range(10):
            actual = spec.get_interbatch_factor(k)
            expected = 1.0
            assert isclose(actual, expected)

    def test_get_interbatch_factor_caches_values(self):
        spec = InstrumentResponseSpec(interbatch_variation=0.8)
        b11 = spec.get_interbatch_factor(1)
        b12 = spec.get_interbatch_factor(1)
        b21 = spec.get_interbatch_factor(2)
        b22 = spec.get_interbatch_factor(2)
        assert b11 == b12
        assert b21 == b22
        assert b11 != b21

    def test_get_sensitivity_loss_cond1(self):
        # max_sensitivity loss set to zero always returns the response factor
        spec = InstrumentResponseSpec()

        # small run order
        actual = spec.get_sensitivity_loss_factor(0)
        expected = 1.0
        assert isclose(actual, expected)

        # large run order
        actual = spec.get_sensitivity_loss_factor(100)
        expected = 1.0
        assert isclose(actual, expected)

    def test_get_sensitivity_loss_cond2(self):
        # max_sensitivity loss set to one, decay set to zero
        spec = InstrumentResponseSpec(max_sensitivity_loss=1.0)

        # zero run order return 1.0
        actual = spec.get_sensitivity_loss_factor(0)
        expected = 1.0
        assert isclose(actual, expected)

        # large run order return 1.0
        actual = spec.get_sensitivity_loss_factor(100)
        expected = 1.0
        assert isclose(actual, expected)

    def test_get_sensitivity_loss_cond3(self):
        # max_sensitivity loss set to zero, non zero decay
        spec = InstrumentResponseSpec(max_sensitivity_loss=0.0, sensitivity_decay=0.5)

        # small run order
        actual = spec.get_sensitivity_loss_factor(0)
        expected = 1.0
        assert isclose(actual, expected)

        # large run order
        actual = spec.get_sensitivity_loss_factor(100)
        expected = 1.0
        assert isclose(actual, expected)

    def test_get_sensitivity_loss_cond4(self):
        # max_sensitivity loss set to one, decay set to zero
        spec = InstrumentResponseSpec(max_sensitivity_loss=1.0, sensitivity_decay=0.5)

        # zero run order return 1.0
        actual = spec.get_sensitivity_loss_factor(0)
        expected = 1.0
        assert isclose(actual, expected)

        # large run should return a smaller value
        actual = spec.get_sensitivity_loss_factor(100)
        assert actual < 0.001



import pytest

from tidyms2.storage.memory import OnMemoryAssayStorage

from ... import helpers


@pytest.fixture
def assay_storage(sample_storage_with_features):
    assay = OnMemoryAssayStorage("assay", helpers.ConcreteRoi, helpers.ConcreteFeature)
    assay.add_sample_data(sample_storage_with_features)
    return assay


class TestAnnotationPatcher:
    @pytest.fixture
    def op(self):
        return helpers.DummyAnnotationPatcher(id="dummy-id")

    def test_apply_ok(self, op, assay_storage: OnMemoryAssayStorage):
        annotations_before = assay_storage.fetch_annotations()
        op.apply(assay_storage)
        annotations_after = assay_storage.fetch_annotations()
        assert annotations_before != annotations_after
        assert all(x.group == op.group for x in annotations_after)


class TestDescriptorPatcher:
    @pytest.fixture
    def op(self):
        return helpers.DummyDescriptorPatcher(id="dummy-id")

    def test_apply_ok(self, op, assay_storage: OnMemoryAssayStorage):
        descriptor = "custom_descriptor"
        descriptors_before = assay_storage.fetch_descriptors(descriptors=[descriptor])
        op.apply(assay_storage)
        descriptors_after = assay_storage.fetch_descriptors(descriptors=[descriptor])
        assert descriptors_before != descriptors_after
        assert all(x == op.patch for x in descriptors_after[descriptor])



import pytest
from sqlalchemy.exc import IntegrityError

from tidyms2.core import exceptions
from tidyms2.core.dataflow import AssayProcessStatus
from tidyms2.core.models import AnnotationPatch, DescriptorPatch, FillValue
from tidyms2.core.utils.common import create_id
from tidyms2.storage import sqlite
from tidyms2.storage.memory import OnMemorySampleStorage
from tidyms2.storage.sqlite.assay import LATEST_SNAPSHOT

from .. import helpers
from ..helpers import ConcreteFeature, ConcreteRoi, create_feature_group


class AssayStorageFixtures:
    @pytest.fixture(scope="class")
    def assay(self):
        return sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)

    @pytest.fixture(scope="class")
    def data_path(self, tmp_path_factory):
        return tmp_path_factory.mktemp("test-on-memory-sample-storage-snapshot")

    @pytest.fixture(scope="class")
    def sample_data1(self, data_path):
        sample = helpers.create_sample(data_path, 1)
        data = OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)
        roi1 = helpers.create_roi(sample)
        roi2 = helpers.create_roi(sample)
        data.add_rois(roi1, roi2)
        ft1 = helpers.create_feature(roi1)
        ft2 = helpers.create_feature(roi2)
        ft1.annotation.group = 0  # type: ignore
        ft2.annotation.group = 0  # type: ignore
        ft3 = helpers.create_feature(roi2)
        data.add_features(ft1, ft2, ft3)
        return data

    @pytest.fixture(scope="class")
    def sample_data2(self, data_path):
        sample = helpers.create_sample(data_path, 2)
        data = OnMemorySampleStorage(sample, helpers.ConcreteRoi, helpers.ConcreteFeature)
        roi1 = helpers.create_roi(sample)
        roi2 = helpers.create_roi(sample)
        data.add_rois(roi1, roi2)
        ft1 = helpers.create_feature(roi1)
        ft2 = helpers.create_feature(roi2)
        ft1.annotation.group = 0  # type: ignore
        ft2.annotation.group = 0  # type: ignore
        ft3 = helpers.create_feature(roi2)
        data.add_features(ft1, ft2, ft3)
        return data

    @pytest.fixture(scope="class")
    def assay_with_sample(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)
        return assay

    @pytest.fixture(scope="class")
    def assay_with_two_samples(self, assay, sample_data1, sample_data2):
        assay.add_sample_data(sample_data1)
        assay.add_sample_data(sample_data2)
        return assay


class TestSQLiteAssayStorageOnDisk(AssayStorageFixtures):
    def test_create_ok(self, sample_data1, tmp_path):
        host = tmp_path / "assay.db"
        assay = sqlite.SQLiteAssayStorage("assay", host, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)
        assert True


class TestSQLiteAssayStorage(AssayStorageFixtures):
    def test_add_sample_with_repeated_id_raises_error(self, assay_with_sample, sample_data1):
        with pytest.raises(exceptions.RepeatedIdError):
            assay_with_sample.add_sample_data(sample_data1)

    def test_fetch_sample(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1):
        expected = sample_data1.get_sample()
        actual = assay_with_sample.fetch_sample(expected.id)
        assert actual == expected

    def test_fetch_sample_non_existing_sample_raises_error(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_sample.fetch_sample("non-existing-sample-id")

    def test_has_sample_existing_sample_returns_true(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1):
        assert assay_with_sample.has_sample(sample_data1.get_sample().id)

    def test_has_sample_non_existing_sample_returns_false(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        assert not assay_with_sample.has_sample("non-existing-sample-id")

    def test_list_samples(self, assay_with_sample, sample_data1):
        expected = [sample_data1.get_sample()]
        actual = assay_with_sample.list_samples()
        assert actual == expected

    def test_get_n_samples_empty_assay_returns_zero(self, assay: sqlite.SQLiteAssayStorage):
        expected = 0
        actual = assay.get_n_samples()
        assert actual == expected

    def test_get_n_samples(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        expected = 1
        actual = assay_with_sample.get_n_samples()
        assert actual == expected

    def test_fetch_rois_by_sample(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1):
        expected = sample_data1.list_rois()
        actual = assay_with_sample.fetch_rois_by_sample(sample_data1.get_sample().id)
        assert actual
        assert all(x.equal(y) for x, y in zip(expected, actual))

    def test_fetch_rois_by_sample_invalid_sample_raise_error(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_sample.fetch_rois_by_sample("invalid-sample-id")

    def test_fetch_rois_by_ids(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1):
        expected = [sample_data1.list_rois()[1]]
        actual = assay_with_sample.fetch_rois_by_id(*(x.id for x in expected))
        assert actual
        assert all(x.equal(y) for x, y in zip(expected, actual))

    def test_fetch_has_roi_true(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1):
        expected_id = sample_data1.list_rois()[0].id
        assert assay_with_sample.has_roi(expected_id)

    def test_fetch_has_roi_false(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        invalid_id = create_id()
        assert not assay_with_sample.has_roi(invalid_id)

    def test_get_n_rois(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        expected = 2
        actual = assay_with_sample.get_n_rois()
        assert actual == expected

    def test_get_n_rois_empty_assay_returns_zero(self, assay: sqlite.SQLiteAssayStorage):
        expected = 0
        actual = assay.get_n_rois()
        assert actual == expected

    def test_fetch_features_by_group(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        group = 0
        actual = assay_with_sample.fetch_features_by_group(group)
        assert len(actual) == 2
        assert all(ft.annotation.group == group for ft in actual)

    def test_fetch_features_by_id(self, assay_with_sample, sample_data1):
        expected = sample_data1.list_features()
        actual = assay_with_sample.fetch_features_by_id(*(x.id for x in expected))
        assert all(x.equal(y) for x, y in zip(actual, expected))

    def test_fetch_features_by_sample_with_shared_roi_share_the_same_roi_instance(
        self, assay_with_sample, sample_data1
    ):
        sample_id = sample_data1.get_sample().id
        ft1, ft2, ft3 = assay_with_sample.fetch_features_by_sample(sample_id)
        assert ft2.roi == ft3.roi
        assert ft2.roi is ft3.roi
        assert ft1.annotation.sample_id == sample_id
        assert ft2.annotation.sample_id == sample_id
        assert ft3.annotation.sample_id == sample_id

    def test_get_n_features_empty_assay_returns_zero(self, assay: sqlite.SQLiteAssayStorage):
        actual = 0
        expected = assay.get_n_features()
        assert actual == expected

    def test_get_n_features(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        actual = 3
        expected = assay_with_sample.get_n_features()
        assert actual == expected

    def test_has_feature_true(self, assay_with_sample: sqlite.SQLiteAssayStorage, sample_data1: OnMemorySampleStorage):
        ft_id = sample_data1.list_features()[1].id
        assert assay_with_sample.has_feature(ft_id)

    def test_has_feature_false(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        ft_id = create_id()
        assert not assay_with_sample.has_feature(ft_id)

    def test_fetch_sample_data_retrieves_equal_data(self, assay_with_sample, sample_data1):
        expected = sample_data1
        sample_id = expected.get_sample().id
        actual = assay_with_sample.fetch_sample_data(sample_id)

        assert actual.list_snapshots() == expected.list_snapshots()
        for snapshot_id in actual.list_snapshots():
            actual.set_snapshot(snapshot_id)
            expected.set_snapshot(snapshot_id)

            assert actual.list_rois() == expected.list_rois()
            assert actual.list_features() == expected.list_features()
            assert actual.get_status() == expected.get_status()
            assert actual.get_sample() == expected.get_sample()


class TestSQLiteSnapshots(AssayStorageFixtures):
    def test_get_process_state(self, assay: sqlite.SQLiteAssayStorage):
        actual = assay.get_process_status()
        expected = AssayProcessStatus()
        assert actual == expected

    def test_set_process_state(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        expected = AssayProcessStatus(feature_matched=True)
        assert not assay.get_process_status() == expected
        assay.set_process_status(expected)
        assert assay.get_process_status() == expected

    def test_create_snapshot(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.create_snapshot("snap")
        assert ["snap", LATEST_SNAPSHOT] == assay.list_snapshots()

    def test_create_snapshot_with_repeated_id_raises_error(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.create_snapshot("snap")
        with pytest.raises(IntegrityError):
            assay.create_snapshot("snap")

    def test_set_snapshot_ok(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        snapshot_id = "snap1"
        assay.create_snapshot(snapshot_id)
        assay.set_snapshot(snapshot_id)
        assert assay.get_snapshot_id() == snapshot_id

    def test_create_snapshot_from_non_latest_snapshot_raises_error(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.create_snapshot("snap1")
        assay.set_snapshot("snap1")
        with pytest.raises(exceptions.SnapshotError):
            assay.create_snapshot("snap2")

    def test_set_snapshot_and_try_to_modify_state_raises_error(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.create_snapshot("snap1")
        assay.set_snapshot("snap1")
        with pytest.raises(exceptions.SnapshotError):
            assay.set_process_status(AssayProcessStatus())

    def test_set_snapshot_and_fetch_annotations(self, assay_with_sample: sqlite.SQLiteAssayStorage):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        annotations_before_set = assay.fetch_annotations()
        assay.create_snapshot("snap1")
        assay.set_snapshot("snap1")
        annotations_after_set = assay.fetch_annotations()
        assert annotations_before_set == annotations_after_set


class TestSQLiteAssayStorageAnnotationApi(AssayStorageFixtures):
    def test_fetch_annotations(self, assay_with_two_samples):
        actual = assay_with_two_samples.fetch_annotations()
        assert len(actual) == assay_with_two_samples.get_n_features()

    def test_fetch_annotations_from_single_sample(self, assay_with_two_samples, sample_data1):
        sample_id = sample_data1.get_sample().id
        actual = assay_with_two_samples.fetch_annotations(sample_id=sample_id)
        assert len(actual) == sample_data1.get_n_features()

    def test_fetch_annotations_from_invalid_sample_raises_error(self, assay_with_two_samples):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_two_samples.fetch_annotations(sample_id="invalid-sample-id")

    def test_patch_annotations_update_data(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)
        annotations_before_patch = assay.fetch_annotations()
        index = 1
        id_ = annotations_before_patch[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        assay.patch_annotations(patch)
        annotations_after_patch = assay.fetch_annotations()
        assert annotations_before_patch != annotations_after_patch
        assert annotations_after_patch[index].group == value

    def test_patch_annotation_and_set_snapshot_previous_snapshot_does_not_apply_patch(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)

        annotations_before_patch = assay.fetch_annotations()
        snapshot_id = "snap"
        assay.create_snapshot(snapshot_id)
        index = 1
        id_ = annotations_before_patch[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        assay.patch_annotations(patch)
        assay.set_snapshot(snapshot_id)
        annotations_from_snapshot = assay.fetch_annotations()
        assert annotations_before_patch == annotations_from_snapshot

    def test_patch_annotation_and_set_snapshot_applies_patch_if_patch_is_from_previous_snapshot(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)

        annotations_before_patch = assay.fetch_annotations()
        snapshot_id = "snap"
        index = 1
        id_ = annotations_before_patch[index].id
        field = "group"
        value = 1000
        patch = AnnotationPatch(id=id_, field=field, value=value)
        assay.patch_annotations(patch)
        assay.create_snapshot(snapshot_id)

        annotations_after_patch = assay.fetch_annotations()

        assay.set_snapshot(snapshot_id)
        annotations_from_snapshot = assay.fetch_annotations()
        assert annotations_before_patch != annotations_after_patch
        assert annotations_after_patch == annotations_from_snapshot


class TestSQLiteAssayStorageDescriptorApi(AssayStorageFixtures):
    def test_fetch_descriptors(self, assay_with_two_samples):
        actual = assay_with_two_samples.fetch_descriptors()
        assert len(actual["mz"]) == assay_with_two_samples.get_n_features()

    def test_fetch_descriptors_from_single_sample(self, assay_with_two_samples, sample_data1):
        sample_id = sample_data1.get_sample().id
        actual = assay_with_two_samples.fetch_descriptors(sample_id=sample_id)
        assert len(actual["mz"]) == sample_data1.get_n_features()

    def test_fetch_annotations_from_invalid_sample_raises_error(self, assay_with_two_samples):
        with pytest.raises(exceptions.SampleNotFound):
            assay_with_two_samples.fetch_descriptors(sample_id="invalid-sample-id")

    def test_patch_descriptors_update_data(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)
        annotations = assay.fetch_annotations()
        descriptors_before_patch = assay.fetch_descriptors()
        index = 1
        id_ = annotations[index].id
        descriptor = "height"
        value = 1000.0
        patch = DescriptorPatch(id=id_, descriptor=descriptor, value=value)
        assay.patch_descriptors(patch)
        descriptors_after_patch = assay.fetch_descriptors()
        assert descriptors_before_patch != descriptors_after_patch
        assert descriptors_after_patch[descriptor][index] == value

    def test_patch_descriptors_and_set_snapshot_previous_snapshot_does_not_apply_patch(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)

        annotations = assay.fetch_annotations()
        descriptors_before_patch = assay.fetch_descriptors()
        snapshot_id = "snap"
        assay.create_snapshot(snapshot_id)
        index = 1
        id_ = annotations[index].id
        descriptor = "height"
        value = 1000.0
        patch = DescriptorPatch(id=id_, descriptor=descriptor, value=value)
        assay.patch_descriptors(patch)
        assay.set_snapshot(snapshot_id)
        descriptors_from_snapshot = assay.fetch_descriptors()
        assert descriptors_before_patch == descriptors_from_snapshot
        assert descriptors_from_snapshot[descriptor][index] != value

    def test_patch_descriptor_and_set_snapshot_applies_patch_if_patch_is_from_previous_snapshot(self, sample_data1):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        assay.add_sample_data(sample_data1)

        annotations = assay.fetch_annotations()
        descriptors_before_patch = assay.fetch_descriptors()
        snapshot_id = "snap"
        index = 1
        id_ = annotations[index].id
        field = "height"
        value = 1000.0
        patch = DescriptorPatch(id=id_, descriptor=field, value=value)
        assay.patch_descriptors(patch)
        assay.create_snapshot(snapshot_id)

        descriptors_after_patch = assay.fetch_descriptors()

        assay.set_snapshot(snapshot_id)
        annotations_from_snapshot = assay.fetch_descriptors()
        assert descriptors_before_patch != descriptors_after_patch
        assert descriptors_after_patch == annotations_from_snapshot


class TestSQLliteAssayStorageFillValueApi(AssayStorageFixtures):
    def test_fetch_fill_values_no_values_return_empty_dict(self, assay):
        fill_values = assay.fetch_fill_values()
        assert isinstance(fill_values, dict)
        assert not fill_values

    def test_fetch_add_values_and_fetch_return_same_fill_values(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        val1 = FillValue(sample_id="sample1", feature_group=0, value=10.0)
        val2 = FillValue(sample_id="sample2", feature_group=1, value=5.0)
        assay.add_fill_values(val1, val2)
        actual = assay.fetch_fill_values()
        assert actual[val1.sample_id][val1.feature_group] == val1.value
        assert actual[val2.sample_id][val2.feature_group] == val2.value

    def test_fetch_fill_values_return_a_copy(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        original = 10.0
        modified = 100.0
        val1 = FillValue(sample_id="sample1", feature_group=0, value=original)
        assay.add_fill_values(val1)
        fill_values_copy1 = assay.fetch_fill_values()
        fill_values_copy1[val1.sample_id][val1.feature_group] = modified
        fill_values_copy2 = assay.fetch_fill_values()
        assert fill_values_copy2[val1.sample_id][val1.feature_group] == original
        assert fill_values_copy1 is not fill_values_copy2

    def test_snapshots_maintain_independent_fill_values_copies(self):
        assay = sqlite.SQLiteAssayStorage("assay", None, ConcreteRoi, ConcreteFeature)
        # add fill values and create a snapshot
        val1 = FillValue(sample_id="sample1", feature_group=0, value=10.0)
        assay.add_fill_values(val1)
        snapshot_id = "snapshot"
        assay.create_snapshot(snapshot_id)

        # add more fill values and set previous snapshot
        val2 = FillValue(sample_id="sample2", feature_group=1, value=5.0)
        assay.add_fill_values(val2)
        assay.set_snapshot(snapshot_id)

        snapshot_fill_values = assay.fetch_fill_values()
        assert snapshot_fill_values[val1.sample_id][val1.feature_group] == val1.value
        assert val2.sample_id not in snapshot_fill_values

        # go back to latest status and check that value 2 is stored
        assay.set_snapshot()
        latest_fill_values = assay.fetch_fill_values()
        assert latest_fill_values[val1.sample_id][val1.feature_group] == val1.value
        assert latest_fill_values[val2.sample_id][val2.feature_group] == val2.value


class TestSQLliteAssayFeatureGroupApi(AssayStorageFixtures):
    def test_fetch_no_groups_return_empty_list(self, assay: sqlite.SQLiteAssayStorage):
        actual = assay.fetch_feature_groups()
        expected = list()
        assert actual == expected

    def tests_add_fetch_return_equal_groups(self, assay: sqlite.SQLiteAssayStorage):
        expected = [create_feature_group(k) for k in range(5)]

        assert not assay.fetch_feature_groups(), "assay must be empty before add"
        assay.add_feature_groups(*expected)
        actual = assay.fetch_feature_groups()

        assert sorted(actual, key=lambda x: x.group) == sorted(expected, key=lambda x: x.group)



"""Tools for feature annotation.

Provides:

1. tools for feature matching (feature group label annotation)
2. Tools for isotopologue annotation.


"""

# from .annotation import annotate, create_annotation_table, create_annotation_tools
# from .isotopologue_annotator import IsotopologueAnnotator

# __all__ = [
#     "annotate",
#     "create_annotation_tools",
#     "create_annotation_table",
#     "IsotopologueAnnotator",
# ]



"""Utilities used by multiple storage classes."""

from typing import Callable

from numpy import array, float64, nanmean, unique, zeros

from ..annotation.consensus import create_consensus_annotation
from ..core.models import Annotation, FeatureGroup, GroupAnnotation, Sample
from ..core.utils.numpy import FloatArray


def create_matrix_data(
    descriptor: list[float],
    annotations: list[Annotation],
    samples: list[Sample],
    fill_values: dict[str, dict[int, float]],
) -> FloatArray:
    """Create the data matrix array."""
    sample_to_index = {s.id: k for k, s in enumerate(samples)}
    n_samples = len(sample_to_index)

    groups = set([x.group for x in annotations])
    groups.discard(-1)

    group_to_index = {g: k for k, g in enumerate(sorted(groups))}
    n_feature_groups = len(group_to_index)

    X = zeros((n_samples, n_feature_groups), dtype=float64)

    for value, ann in zip(descriptor, annotations):
        if ann.group > -1:
            row = sample_to_index[ann.sample_id]
            col = group_to_index[ann.group]
            X[row, col] = value

    for sample_id, sample_fills in fill_values.items():
        for group, value in sample_fills.items():
            row = sample_to_index[sample_id]
            col = group_to_index[group]
            X[row, col] = value

    return X


def create_feature_groups(
    descriptors: dict[str, list[float]],
    annotations: list[Annotation],
    agg: dict[str, Callable] | Callable | None = None,
) -> list[FeatureGroup]:
    """Compute feature groups from a feature list."""
    n_ft_descriptors = len(next(iter(descriptors.values())))
    assert n_ft_descriptors == len(annotations), "incompatible descriptor and annotation length."

    if isinstance(agg, dict):
        agg = agg.update({k: nanmean for k in descriptors if k not in agg})
    else:
        default_agg = agg or nanmean
        agg = {k: default_agg for k in descriptors}

    group_annotations = create_consensus_annotation(annotations)

    group_labels, group_index = unique(array([x.group for x in annotations], dtype="int64"), return_index=True)
    descriptor_array = {k: array(v) for k, v in descriptors.items()}
    features_groups = list()
    for label, index in zip(group_labels, group_index):
        if label == -1:
            continue

        assert agg is not None  # helping pyright...
        group_descriptors = {k: agg[k](v[index]).item() for k, v in descriptor_array.items()}
        ann = group_annotations.get(label) or GroupAnnotation(label=label)
        group = FeatureGroup(group=label, annotation=ann, descriptors=group_descriptors)
        features_groups.append(group)

    return features_groups



from pathlib import Path
from random import seed

import pytest
from numpy import allclose, nan
from pydantic import ValidationError

from tidyms2 import operators
from tidyms2.core.enums import AggregationMethod, SampleType
from tidyms2.core.exceptions import EmptyDataMatrix, ProcessStatusError
from tidyms2.core.models import Sample
from tidyms2.core.operators.pipeline import Pipeline
from tidyms2.simulation.lcms import SimulatedLCMSAdductSpec, simulate_data_matrix
from tidyms2.simulation.utils import create_sample_list


class TestBlankCorrector:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "blank" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 100.0},
                    "group-b": {"mean": 1000.0},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 100.0},
                    "group-b": {"mean": 1000.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_default(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        op = operators.BlankCorrector()
        op.apply(data)

        sample_ids = data.query.filter(type="sample").fetch_sample_ids()[0][1]
        blank_sample_ids = data.query.filter(type="blank").fetch_sample_ids()[0][1]
        assert allclose(data.get_data(sample_ids=sample_ids), 900.0)
        assert allclose(data.get_data(sample_ids=blank_sample_ids), 100.0)

    def test_with_blank_groups(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        op = operators.BlankCorrector(blank_groups=["group-a"])
        op.apply(data)

        sample_ids = data.query.filter(type="sample").fetch_sample_ids()[0][1]
        blank_sample_ids = data.query.filter(type="blank").fetch_sample_ids()[0][1]
        assert allclose(data.get_data(sample_ids=sample_ids), 900.0)
        assert allclose(data.get_data(sample_ids=blank_sample_ids), 100.0)

    def test_with_excluded_samples(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        excluded = data.samples[2].id
        row_before = data.get_rows(excluded)[0]
        op = operators.BlankCorrector(exclude_samples=[excluded])
        op.apply(data)

        row_after = data.get_rows(excluded)[0]
        assert allclose(row_before.data, row_after.data)

    def test_with_excluded_features(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        col_before = data.get_columns(1)[0]
        op = operators.BlankCorrector(exclude_features=[1])
        op.apply(data)
        col_after = data.get_columns(1)[0]
        assert allclose(col_before.data, col_after.data)

    def test_with_excluded_features_and_samples(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        excluded_sample = data.samples[2].id
        col_before = data.get_columns(1)[0]
        row_before = data.get_rows(excluded_sample)[0]
        op = operators.BlankCorrector(exclude_features=[1], exclude_samples=[excluded_sample])
        op.apply(data)
        col_after = data.get_columns(1)[0]
        row_after = data.get_rows(excluded_sample)[0]
        assert allclose(col_before.data, col_after.data)
        assert allclose(row_before.data, row_after.data)

    def test_default_correct_blanks(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        op = operators.BlankCorrector(apply_to_blanks=True)
        op.apply(data)

        sample_ids = data.query.filter(type="sample").fetch_sample_ids()[0][1]
        blank_sample_ids = data.query.filter(type="blank").fetch_sample_ids()[0][1]
        assert allclose(data.get_data(sample_ids=sample_ids), 900.0)
        assert allclose(data.get_data(sample_ids=blank_sample_ids), 0.0)

    def test_default_no_blank_raise_error(self, adducts, samples):
        for sample in samples:
            sample.meta.type = "sample"
        data = simulate_data_matrix(adducts, samples)
        op = operators.BlankCorrector()
        with pytest.raises(ValueError):
            op.apply(data)


class TestSampleFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "qc" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 1000.0},
                    "group-b": {"mean": 1000.0, "std": 1000.0},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 1000.0, "std": 1000.0},
                    "group-b": {"mean": 1000.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_remove_sample_group(self, samples, adducts):
        data = simulate_data_matrix(adducts, samples)
        op = operators.SampleFilter(group="group-a")  # type: ignore
        op.apply(data)
        assert data.get_n_samples()
        assert all(sample.meta.group == "group-b" for sample in data.samples)

    def test_remove_sample_type(self, samples, adducts):
        data = simulate_data_matrix(adducts, samples)
        op = operators.SampleFilter(type="qc")  # type: ignore
        op.apply(data)
        assert data.get_n_samples()
        assert all(sample.meta.type == "sample" for sample in data.samples)


class TestFeatureFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "qc" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "rt": {"mean": 50.0},
                "abundance": {
                    "group-a": {"mean": 1000.0},
                    "group-b": {"mean": 1000.0, "std": 1000.0},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "rt": {"mean": 100.0},
                "abundance": {
                    "group-a": {"mean": 1000.0, "std": 1000.0},
                    "group-b": {"mean": 1000.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_filter_is_not_a_pair_raises_error(self):
        with pytest.raises(ValidationError):
            operators.FeatureFilter(rt=125)  # type: ignore

    def test_remove_features_using_rt(self, samples, adducts):
        data = simulate_data_matrix(adducts, samples)
        op = operators.FeatureFilter(rt=(75, 125))  # type: ignore
        op.apply(data)
        assert data.get_n_features() == 1
        assert data.has_feature(1)


class TestCVFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "qc" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 1000.0},
                    "group-b": {"mean": 1000.0, "std": 200.0},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 1000.0, "std": 200.0},
                    "group-b": {"mean": 1000.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_lb_greater_than_ub_raises_error(self):
        with pytest.raises(ValidationError):
            operators.CVFilter(lb=0.5, ub=0.25)

    def test_default(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # only adduct 2 should be removed as it has CV ~ 1.0 in the QC type
        rm_group = data.features[1].group
        assert data.get_n_features() == 2
        assert data.has_feature(rm_group)
        op = operators.CVFilter(ub=0.01)
        op.apply(data)
        assert data.get_n_features() == 1
        assert not data.has_feature(rm_group)

    def test_default_no_qc_type_raise_error(self, adducts, samples):
        for sample in samples:
            sample.meta.type = "sample"
        data = simulate_data_matrix(adducts, samples)
        op = operators.CVFilter()

        with pytest.raises(ValueError):
            op.apply(data)

    def test_using_filter(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # only adduct 1 should be removed as it has CV ~ 1.0 in the group-b
        rm_group = data.features[0].group
        assert data.get_n_features() == 2
        assert data.has_feature(rm_group)
        op = operators.CVFilter(ub=0.01, filter={"group": "group-b"})
        op.apply(data)
        assert data.get_n_features() == 1
        assert not data.has_feature(rm_group)

    def test_group_by(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # using max aggregation should remove both features as both have a
        # a group with CV ~ 1.0
        op = operators.CVFilter(ub=0.01, group_by=["group"], aggregation=AggregationMethod.MIN)
        op.apply(data)
        assert data.get_n_features() == len(adducts)

    def test_aggregation(self, adducts, samples):
        seed(1000)
        data = simulate_data_matrix(adducts, samples)
        # using max aggregation should remove both features as both have a
        # a group with CV ~ 1.0
        op = operators.CVFilter(ub=0.01, group_by=["group"], aggregation=AggregationMethod.MAX)
        with pytest.raises(EmptyDataMatrix):
            op.apply(data)


class TestDRatioFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(20):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "qc" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 2000.0, "std": 100.0},  # QC
                    "group-b": {"mean": 2000.0, "std": 10.0},  # sample
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 2000.0, "std": 10.0},  # QC
                    "group-b": {"mean": 2000.0, "std": 100.0},  # sample
                },
            }
        )
        return [adduct1, adduct2]

    def test_using_filter_param_raises_error(self):
        with pytest.raises(ValidationError):
            operators.DRatioFilter(filter={"type": "sample"})

    def test_using_group_by_param_raises_error(self):
        with pytest.raises(ValidationError):
            operators.DRatioFilter(group_by=["type"])

    def test_default(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # only adduct 1 should be removed as its D-ratio is > 1.0
        rm_group = data.features[0].group
        assert data.get_n_features() == 2
        assert data.has_feature(rm_group)
        op = operators.DRatioFilter()
        op.apply(data)
        assert data.get_n_features() == 1
        assert not data.has_feature(rm_group)

    def test_default_no_qc_type_raise_error(self, adducts, samples):
        for sample in samples:
            sample.meta.type = "sample"
        data = simulate_data_matrix(adducts, samples)
        op = operators.DRatioFilter()

        with pytest.raises(ValueError):
            op.apply(data)

    def test_default_no_sample_type_raise_error(self, adducts, samples):
        for sample in samples:
            sample.meta.type = "qc"
        data = simulate_data_matrix(adducts, samples)
        op = operators.DRatioFilter()

        with pytest.raises(ValueError):
            op.apply(data)


class TestDetectionRateFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "blank" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 100.0},
                    "group-b": {"mean": 1000.0, "prevalence": 0.5},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 100.0},
                    "group-b": {"prevalence": 1.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_default(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # only adduct 1 should be removed as it has DR ~ 0.5 in the sample sample type
        rm_group = data.features[0].group
        assert data.get_n_features() == 2
        assert data.has_feature(rm_group)
        op = operators.DetectionRateFilter()
        op.apply(data)
        assert data.get_n_features() == 1
        assert not data.has_feature(rm_group)

    def test_default_no_blank(self, adducts, samples):
        for sample in samples:
            sample.meta.type = "sample"
        data = simulate_data_matrix(adducts, samples)
        op = operators.DetectionRateFilter()
        op.apply(data)
        assert data.get_n_features() == 1


class TestCorrelationFilter:
    @pytest.fixture
    def samples(self):
        sample_list = list()
        for k in range(10):
            id_ = f"sample-{k}"
            path = Path.cwd() / f"{id_}.mzML"
            s = Sample(id=id_, path=path)
            s.meta.order = k
            s.meta.group = "group-a" if k % 2 else "group-b"
            s.meta.type = "blank" if k % 2 else "sample"
            sample_list.append(s)
        return sample_list

    @pytest.fixture
    def adducts(self):
        adduct1 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "group-a": {"mean": 100.0, "std": 1.0},
                    "group-b": {"mean": 100.0, "std": 1.0},
                },
            },
        )

        adduct2 = SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "group-a": {"mean": 100.0, "std": 1.0},
                    "group-b": {"mean": 100.0, "std": 1.0},
                },
            }
        )
        return [adduct1, adduct2]

    def test_data_with_missing_values_raise_error(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        data_with_nan = data.get_data()
        data_with_nan[1] = nan
        data.set_data(data_with_nan)
        op = operators.CorrelationFilter(field="order", lb=1, ub=2)
        with pytest.raises(ProcessStatusError):
            op.apply(data)

    def test_default(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        # should remove both features as max corr is 1.0
        op = operators.CorrelationFilter(field="order", lb=1, ub=2)
        with pytest.raises(EmptyDataMatrix):
            op.apply(data)

    def test_default_features_with_correlation(self, adducts, samples):
        data = simulate_data_matrix(adducts, samples)
        order = data.list_sample_field("order")
        data_with_corr = data.get_data()
        data_with_corr[:, 0] = order
        data_with_corr[:, 1] = -1 * data_with_corr[:, 0]
        data.set_data(data_with_corr)

        op = operators.CorrelationFilter(field="order", lb=0.5, ub=1.0)
        op.apply(data)

        assert data.has_feature(0)
        assert not data.has_feature(1)


def test_matrix_processing_pipeline():
    sample_types = [
        SampleType.BLANK,
        SampleType.TECHNICAL_QC,
        SampleType.SAMPLE,
        SampleType.SAMPLE,
        SampleType.TECHNICAL_QC,
    ]

    samples = create_sample_list(sample_types)
    adducts = [
        SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C54H104O6]+",
                "abundance": {
                    "blank": {"mean": 200.0},
                    "sample": {"mean": 100.0, "prevalence": 1.0},
                    "qc": {"mean": 1000.0, "prevalence": 1.0},
                },
            },
        ),
        SimulatedLCMSAdductSpec.model_validate(
            {
                "formula": "[C27H40O2]+",
                "abundance": {
                    "blank": {"mean": 10.0},
                    "sample": {"mean": 1000.0, "prevalence": 1.0},
                    "qc": {"mean": 1000.0, "prevalence": 1.0},
                },
            }
        ),
    ]
    matrix = simulate_data_matrix(adducts, samples)
    pipe = Pipeline("matrix-pipe")
    pipe.add_operator(operators.BlankCorrector(id="blank-corrector"))
    pipe.add_operator(operators.DetectionRateFilter(group_by=["group"], threshold=10.0))
    pipe.apply(matrix)
    # first feature is always set to zero after blank correction,
    # so it should be removed by the detection rate filter
    assert matrix.get_n_features() == 1



from tidyms2.core.enums import MSInstrument, Polarity, SeparationMode
from tidyms2.lcms import create_lcms_assay


def test_e2e_lcms_assay_on_memory_storage_sequential_sample_executor(lcms_sample_factory):
    assay = create_lcms_assay(
        "test-lcms-assay-e2e",
        instrument=MSInstrument.QTOF,
        separation=SeparationMode.UPLC,
        polarity=Polarity.POSITIVE,
        annotate_isotopologues=True,
    )
    n_samples = 10
    samples = (lcms_sample_factory(f"sample-{k}", order=k) for k in range(n_samples))
    assay.add_samples(*samples)

    assay.process_samples()
    assay.process_assay()
    matrix = assay.create_data_matrix()

    assert matrix.get_n_samples() == n_samples
    assert matrix.get_n_features() == 6  # see lcms_sample_factory on conftest


def test_e2e_lcms_assay_on_disk_storage_sequential_sample_executor(lcms_sample_factory, tmp_path):
    assay = create_lcms_assay(
        "test-lcms-assay-e2e",
        instrument=MSInstrument.QTOF,
        separation=SeparationMode.UPLC,
        polarity=Polarity.POSITIVE,
        annotate_isotopologues=True,
        on_disk=True,
        storage_path=str(tmp_path / "data.db"),
    )
    n_samples = 10
    samples = (lcms_sample_factory(f"sample-{k}", order=k) for k in range(n_samples))
    assay.add_samples(*samples)

    assay.process_samples()
    assay.process_assay()

    matrix = assay.create_data_matrix()

    assert matrix.get_n_samples() == n_samples
    assert matrix.get_n_features() == 6  # see lcms_sample_factory on conftest


def test_e2e_lcms_assay_on_memory_storage_parallel_sample_executor(lcms_sample_factory):
    assay = create_lcms_assay(
        "test-lcms-assay-e2e",
        instrument=MSInstrument.QTOF,
        separation=SeparationMode.UPLC,
        polarity=Polarity.POSITIVE,
        annotate_isotopologues=True,
        max_workers=2,
    )
    n_samples = 10
    samples = (lcms_sample_factory(f"sample-{k}", order=k) for k in range(n_samples))
    assay.add_samples(*samples)

    assay.process_samples()
    assay.process_assay()

    matrix = assay.create_data_matrix()

    assert matrix.get_n_samples() == n_samples
    assert matrix.get_n_features() == 6  # see lcms_sample_factory on conftest


def test_e2e_lcms_assay_on_disk_storage_parallel_sample_executor(lcms_sample_factory, tmp_path):
    assay = create_lcms_assay(
        "test-lcms-assay-e2e",
        instrument=MSInstrument.QTOF,
        separation=SeparationMode.UPLC,
        polarity=Polarity.POSITIVE,
        annotate_isotopologues=True,
        on_disk=True,
        max_workers=2,
        storage_path=str(tmp_path / "data.db"),
    )
    n_samples = 10
    samples = (lcms_sample_factory(f"sample-{k}", order=k) for k in range(n_samples))
    assay.add_samples(*samples)

    assay.process_samples()
    assay.process_assay()

    matrix = assay.create_data_matrix()

    assert matrix.get_n_samples() == n_samples
    assert matrix.get_n_features() == 6  # see lcms_sample_factory on conftest



"""Raw data reader interface."""

from __future__ import annotations

import pathlib
from typing import Protocol

from ..core.models import Chromatogram, MSSpectrum, Sample
from ..core.registry import Registry

reader_registry: Registry[Reader] = Registry("reader")


class Reader(Protocol):
    """Reader interface for raw data."""

    def __init__(self, src: pathlib.Path | Sample): ...

    def get_chromatogram(self, index: int) -> Chromatogram:
        """Retrieve a chromatogram from file."""
        ...

    def get_spectrum(self, index: int) -> MSSpectrum:
        """Retrieve a spectrum from file."""
        ...

    def get_n_chromatograms(self) -> int:
        """Retrieve the total number of chromatogram."""
        ...

    def get_n_spectra(self) -> int:
        """Retrieve the total number of spectra."""
        ...



"""In memory sample data storage implementation."""

from __future__ import annotations

from typing import Generic, Iterable
from uuid import UUID

from ..core import exceptions, storage
from ..core.dataflow import AssayProcessStatus, SampleProcessStatus
from ..core.models import (
    Annotation,
    AnnotationPatch,
    DescriptorPatch,
    FeatureGroup,
    FeatureType,
    FillValue,
    RoiType,
    Sample,
)
from ..core.storage import SampleStorage
from ..core.utils.common import create_id
from .table import FeatureTable

LATEST = "head"


class OnMemoryAssayStorage(Generic[RoiType, FeatureType]):
    """Store assay data in memory."""

    def __init__(self, id: str, roi_type: type[RoiType], feature_type: type[FeatureType]) -> None:
        self.id = id

        self._feature_type = feature_type
        self._roi_type = roi_type

        self._sample_data: dict[str, OnMemorySampleStorage[RoiType, FeatureType]] = dict()
        self._rois_to_sample_id: dict[UUID, str] = dict()

        self._current = OnMemoryAssayStorageSnapshot(LATEST, AssayProcessStatus())
        self._snapshots = [self._current]

    def add_sample_data(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        """Add samples to the assay."""
        sample = data.get_sample()
        if self.has_sample(sample.id):
            raise exceptions.RepeatedIdError(sample.id)

        copy = OnMemorySampleStorage.from_sample_storage(data)

        new_rois = dict()
        sample_id = copy.get_sample().id
        new_rois.update({x.id: sample_id for x in copy.list_rois()})

        new_features = dict()
        sample_id = copy.get_sample().id
        features = copy.list_features()
        new_features.update({x.id: sample_id for x in features})

        descriptors = list()
        annotations = list()
        for ft in features:
            descriptors.append(ft.describe())
            annotations.append(ft.annotation)

        annotations = [x.annotation for x in features]
        self._current.add_descriptors(descriptors, annotations)

        self._rois_to_sample_id.update(new_rois)
        self._sample_data[sample.id] = copy
        self._groups = list()

    def add_fill_values(self, *fill_values: FillValue) -> None:
        """Add values to fill missing data matrix entries."""
        d = dict()
        for fill in fill_values:
            sample_d = d.setdefault(fill.sample_id, dict())
            sample_d[fill.feature_group] = fill.value
        self._current.add_fill_values(d)

    def add_feature_groups(self, *feature_groups: FeatureGroup) -> None:
        """Add feature groups to the assay."""
        self._current.add_feature_groups(*feature_groups)

    def fetch_feature_groups(self) -> list[FeatureGroup]:
        """Fetch feature groups from the assay."""
        return self._current.fetch_feature_groups()

    def create_snapshot(self, snapshot_id) -> None:
        """Create a new sample data snapshot.

        :param snapshot_id: the id for the new snapshot.
        :raises RepeatedIdError: if a snapshot with this id already exists.

        """
        self._check_latest_snapshot()
        if snapshot_id == LATEST:
            raise ValueError(f"snapshot id `{LATEST}` is reserved for internal use.")

        if snapshot_id in self.list_snapshots():
            msg = f"Snapshot with id={snapshot_id} already exists for assay {self.id}."
            raise exceptions.RepeatedIdError(msg)

        self._current.id = snapshot_id
        latest = self._current.copy_snapshot(LATEST)
        self._snapshots.append(latest)
        self._current = latest

    def fetch_fill_values(self) -> dict[str, dict[int, float]]:
        """Fetch fill values for missing data matrix entries."""
        return self._current.fetch_fill_values(copy=True)

    def get_process_status(self) -> AssayProcessStatus:
        """Retrieve the current process status."""
        return self._current.status.model_copy(deep=True)

    def set_process_status(self, status: AssayProcessStatus) -> None:
        """Set the assay process status."""
        self._current.status = status

    def fetch_sample_data(self, sample_id: str) -> OnMemorySampleStorage[RoiType, FeatureType]:
        """Fetch Samples from the assay using their ids."""
        if not self.has_sample(sample_id):
            raise exceptions.SampleNotFound(sample_id)
        return self._sample_data[sample_id]

    def fetch_sample(self, sample_id: str) -> Sample:
        """Fetch Samples from the assay using their ids."""
        return self.fetch_sample_data(sample_id).get_sample()

    def has_sample(self, sample_id: str) -> bool:
        """Check if the assay contains a sample with the provided id."""
        return sample_id in self._sample_data

    def list_samples(self) -> list[Sample]:
        """Fetch all samples in the assay."""
        return [x.get_sample().model_copy(deep=True) for x in self._sample_data.values()]

    def list_snapshots(self) -> list[str]:
        """List all snapshot ids."""
        return [x.id for x in self._snapshots]

    def list_feature_groups(self) -> list[int]:
        """List all feature groups in the assay."""
        return self._current.list_feature_groups()

    def get_feature_type(self) -> type[FeatureType]:
        """Retrieve the Feature class used."""
        return self._feature_type

    def get_roi_type(self) -> type[RoiType]:
        """Retrieve the ROI class used."""
        return self._roi_type

    def get_n_features(self) -> int:
        """Get the total number of features in the assay."""
        return sum(x.get_n_features() for x in self._sample_data.values())

    def get_n_rois(self) -> int:
        """Get the total number of ROIs in the assay."""
        return sum(x.get_n_rois() for x in self._sample_data.values())

    def get_snapshot_id(self) -> str:
        """Get the current snapshot id."""
        return self._current.id

    def has_roi(self, roi_id: UUID) -> bool:
        """Check if a ROI is in the storage."""
        return roi_id in self._rois_to_sample_id

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a Feature with the provided id is in the storage."""
        return self._current.has_feature(feature_id)

    def has_feature_group(self, feature_group: int) -> bool:
        """Check if a group with the provided id is in the assay."""
        return self._current.has_feature_group(feature_group)

    def fetch_annotations(self, sample_id: str | None = None) -> list[Annotation]:
        """Fetch a copy of the feature annotations.

        :param sample_id: If provided, only fetch annotations from this sample. By default, fetch annotations
            from all samples.
        :raises SampleNotFound: if a sample id that is not in the assay storage is provided.

        """
        if sample_id is not None and not self.has_sample(sample_id):
            raise exceptions.SampleNotFound(sample_id)

        return self._current.fetch_annotations(sample_id=sample_id, copy=True)

    def fetch_descriptors(
        self, sample_id: str | None = None, descriptors: Iterable[str] | None = None
    ) -> dict[str, list[float]]:
        """Fetch a copy of the feature descriptors.

        :param sample_id: If provided, only fetch descriptors from this sample. Otherwise, fetch descriptors
            from all samples
        :param descriptors: If provided only fetch values from these descriptors. By default, all descriptors
            are fetched.
        :raises SampleNotFound: if a sample id that is not in the assay storage is provided.
        :raises InvalidFeatureDescriptor: If an undefined descriptor name for the assay feature type is provided.
        """
        if sample_id is not None and not self.has_sample(sample_id):
            raise exceptions.SampleNotFound(sample_id)

        all_descriptors = self._feature_type.descriptor_names()
        if descriptors is None:
            descriptors = list(all_descriptors)

        for d in descriptors:
            if d not in all_descriptors:
                msg = f"{d} is not a valid descriptor of {self._feature_type.__name__}."
                raise exceptions.InvalidFeatureDescriptor(msg)

        return self._current.fetch_descriptors(descriptors=descriptors, sample_id=sample_id, copy=True)

    def fetch_rois_by_sample(self, sample_id: str) -> list[RoiType]:
        """Retrieve ROIs from the storage."""
        return [x.model_copy(deep=True) for x in self.fetch_sample_data(sample_id).list_rois()]

    def fetch_rois_by_id(self, *roi_ids: UUID) -> list[RoiType]:
        """Fetch a ROI using its id."""
        roi_list = list()
        for id_ in roi_ids:
            if not self.has_roi(id_):
                raise exceptions.RoiNotFound(id_)

            sample_id = self._rois_to_sample_id[id_]
            roi = self._sample_data[sample_id].get_roi(id_).model_copy(deep=True)
            roi_list.append(roi)
        return roi_list

    def fetch_features_by_id(self, *feature_ids: UUID) -> list[FeatureType]:
        """Fetch a feature using its id."""
        # first we list all features by id
        feature_list = list()
        for id_ in feature_ids:
            if not self.has_feature(id_):
                raise exceptions.FeatureNotFound(id_)
            sample_id = self._current.get_sample_id(id_)
            ft = self._sample_data[sample_id].get_feature(id_)
            feature_list.append(ft)

        # multiple features may be extracted from the same ROI, we create a copy of each
        # unique ROI to avoid duplicates
        unique_roi_ids = {x.roi.id for x in feature_list}
        roi_id_to_roi = {x.id: x for x in self.fetch_rois_by_id(*unique_roi_ids)}

        # finally we create copy of all features but pass the same ROI copy to features
        # that share ROIs
        return [ft.model_copy(deep=True, update={"roi": roi_id_to_roi[ft.roi.id]}) for ft in feature_list]

    def fetch_features_by_sample(self, sample_id: str) -> list[FeatureType]:
        """Retrieve all features from a sample."""
        # here we also have the problem of shared ROIs as in the fetch_features_by_id method
        # so we use the same approach
        all_ids = [x.id for x in self.fetch_sample_data(sample_id).list_features()]
        return self.fetch_features_by_id(*all_ids)

    def fetch_features_by_group(self, group: int) -> list[FeatureType]:
        """Retrieve all features belonging to a feature group."""
        return self.fetch_features_by_id(*self._current.get_ids_by_group(group))

    def patch_annotations(self, *patches: AnnotationPatch) -> None:
        """Update feature annotation values."""
        self._check_latest_snapshot()
        self._current.patch_annotation(*patches)

    def patch_descriptors(self, *patches: DescriptorPatch) -> None:
        """Update feature descriptor values."""
        self._check_latest_snapshot()
        self._current.patch_descriptors(*patches)

    def set_snapshot(self, snapshot_id: str | None = None, reset: bool = False) -> None:
        """Set snapshot from which the storage will fetch data from.

        :param snapshot_id: the snapshot to set
        :param reset: set the selected snapshot as the latest and delete posterior snapshots.
            Note that the selected snapshot id will be set to `head`.
        :raises SnapshotNotFoundError: if the provided `snapshot_id` is not in the storage

        """
        if snapshot_id is None:
            snapshot_id = LATEST
        try:
            snapshot_index = self.list_snapshots().index(snapshot_id)
            self._current = self._snapshots[snapshot_index]
        except ValueError as e:
            msg = f"Snapshot {snapshot_id} not found in {self.id} assay storage."
            raise exceptions.SnapshotNotFound(msg) from e

        if reset:
            self._snapshots = self._snapshots[:snapshot_index]
            self._current.id = LATEST

    def _check_latest_snapshot(self):
        if self._current.id != LATEST:
            msg = f"Only latest snapshot can be modified. Cannot change the state of snapshot {self._current.id}"
            raise exceptions.SnapshotError(msg)


@storage.register_sample_storage
class OnMemorySampleStorage(Generic[RoiType, FeatureType]):
    """Store sample data in memory.

    Manages accession to sample and ROIs in O(1) time. Both add features and
    add ROI operations are atomic and consistent operations.

    """

    def __init__(self, sample: Sample, roi_type: type[RoiType], feature_type: type[FeatureType], **kwargs) -> None:
        self.roi_type = roi_type
        self.feature_type = feature_type
        self._sample = sample.model_copy()
        self._snapshots = list()
        self._current: None | OnMemorySampleStorageSnapshot = None

    def _add_snapshot(self, snapshot: OnMemorySampleStorageSnapshot[RoiType, FeatureType]):
        self._snapshots.append(snapshot)

    def _get_current_snapshot(self) -> OnMemorySampleStorageSnapshot[RoiType, FeatureType]:
        # create a snapshot if it does not exist
        if self._current is None:
            state = SampleProcessStatus()
            latest = OnMemorySampleStorageSnapshot(self._sample, LATEST, state, self.roi_type, self.feature_type)
            self._current = latest
            self._snapshots.append(latest)
        return self._current

    def add_features(self, *features: FeatureType) -> None:
        """Add features to the sample storage.

        :param features: the features to be add.
        :raises RepeatedIdError: if a feature with an existing id is provided.
        :raises RoiNotFoundError: if trying to add a feature associated with a ROI not in the storage

        """
        self._check_latest_snapshot()
        self._get_current_snapshot().add_features(*features)

    def add_rois(self, *rois: RoiType) -> None:
        """Add ROIs to the sample storage.

        :param rois: the rois to be add
        :raises RepeatedIdError: if a ROI with this id already exists.

        """
        self._check_latest_snapshot()
        self._get_current_snapshot().add_rois(*rois)

    def create_snapshot(self, snapshot_id: str) -> None:
        """Create a new sample data snapshot.

        :param snapshot_id: the id for the new snapshot.
        :raises RepeatedIdError: if a snapshot with this id already exists.

        """
        self._check_latest_snapshot()
        if snapshot_id == LATEST:
            raise ValueError(f"snapshot id `{LATEST}` is reserved for internal use.")

        if snapshot_id in self.list_snapshots():
            msg = f"Snapshot with id={snapshot_id} already exists for sample {self._sample.id}."
            raise exceptions.RepeatedIdError(msg)

        self._snapshots[-1].id = snapshot_id

        latest = self._get_current_snapshot().copy(LATEST, set_new_ids=True)

        self._snapshots.append(latest)
        self._current = latest

    def delete_features(self, *feature_ids: UUID) -> None:
        """Delete features using their ids.

        Non-existing ids are ignored.

        """
        self._check_latest_snapshot()
        self._get_current_snapshot().delete_features(*feature_ids)

    def delete_rois(self, *roi_ids: UUID) -> None:
        """Delete ROIs using their ids.

        Non-existing ids are ignored.

        """
        self._check_latest_snapshot()
        self._get_current_snapshot().delete_rois(*roi_ids)

    def get_feature(self, feature_id: UUID) -> FeatureType:
        """Retrieve a feature by id.

        :raises FeatureNotFoundError: if the provided `feature_id` is not in the storage

        """
        return self._get_current_snapshot().get_feature(feature_id)

    def get_n_features(self) -> int:
        """Get the total number of features in the storage."""
        return self._get_current_snapshot().get_n_features()

    def get_n_rois(self) -> int:
        """Get the total number of ROIs in the storage."""
        return self._get_current_snapshot().get_n_rois()

    def get_roi(self, roi_id: UUID) -> RoiType:
        """Retrieve a ROI by id.

        :raises RoiNotFoundError: if the provided `roi_id` is not in the storage

        """
        return self._get_current_snapshot().get_roi(roi_id)

    def get_sample(self) -> Sample:
        """Retrieve the storage sample."""
        return self._sample

    def get_snapshot_id(self) -> str:
        """Get the current snapshot id."""
        return self._get_current_snapshot().id

    def get_status(self) -> SampleProcessStatus:
        """Get the current process status."""
        return self._get_current_snapshot().status

    def has_feature(self, feature_id: UUID) -> bool:
        """Check the existence of a feature using its id."""
        return self._get_current_snapshot().has_feature(feature_id)

    def has_roi(self, roi_id: UUID) -> bool:
        """Check the existence of a ROI with the specified id."""
        return self._get_current_snapshot().has_roi(roi_id)

    def list_features(self, roi_id: UUID | None = None) -> list[FeatureType]:
        """List stored features.

        :param roi_id: if provided, only features associated with this ROI are listed
        :raises RoiNotFoundError: if the provided `roi_id` is not in the storage

        """
        return self._get_current_snapshot().list_features(roi_id)

    def list_rois(self) -> list[RoiType]:
        """List all stored ROIs."""
        return self._get_current_snapshot().list_rois()

    def list_snapshots(self) -> list[str]:
        """List all snapshots."""
        return [x.id for x in self._snapshots]

    def set_snapshot(self, snapshot_id: str | None = None, reset: bool = False) -> None:
        """Set snapshot from which the storage will fetch data from.

        :param snapshot_id: the snapshot to set
        :param reset: set the selected snapshot as the latest and delete posterior snapshots.
            The selected snapshot id will be set to `head`.
        :raises SnapshotNotFoundError: if the provided `snapshot_id` is not in the storage

        """
        if snapshot_id is None:
            snapshot_id = LATEST
        try:
            snapshot_index = self.list_snapshots().index(snapshot_id)
            self._current = self._snapshots[snapshot_index]
        except ValueError as e:
            msg = f"Snapshot {snapshot_id} not found in {self._sample.id} storage."
            raise exceptions.SnapshotNotFound(msg) from e

        if reset:
            self._snapshots = self._snapshots[:snapshot_index]
            self._get_current_snapshot().id = LATEST

    def set_status(self, status: SampleProcessStatus) -> None:
        """Set the current process status."""
        self._check_latest_snapshot()
        self._get_current_snapshot().status = status

    def _check_latest_snapshot(self):
        id_ = self._get_current_snapshot().id
        if id_ != LATEST:
            msg = f"Only latest snapshot can be modified. Cannot change the state of snapshot {id_}"
            raise exceptions.SnapshotError(msg)

    @classmethod
    def from_sample_storage(
        cls, sample_storage: SampleStorage[RoiType, FeatureType]
    ) -> OnMemorySampleStorage[RoiType, FeatureType]:
        """Create a new instance using the provided sample storage."""
        sample = sample_storage.get_sample()
        copied = OnMemorySampleStorage(sample, sample_storage.roi_type, sample_storage.feature_type)

        current_snapshot_id = sample_storage.get_snapshot_id()

        for snapshot_id in sample_storage.list_snapshots():
            sample_storage.set_snapshot(snapshot_id)
            snapshot = OnMemorySampleStorageSnapshot.from_sample_storage(sample_storage)
            copied._add_snapshot(snapshot)
        copied.set_snapshot(LATEST)

        sample_storage.set_snapshot(current_snapshot_id)
        return copied

    @classmethod
    def from_dict(
        cls,
        sample: Sample,
        rois: dict[str, list[RoiType]],
        features: dict[str, list[FeatureType]],
        snapshots: list[str],
        states: dict[str, SampleProcessStatus],
        roi_type: type[RoiType],
        feature_type: type[FeatureType],
    ) -> OnMemorySampleStorage[RoiType, FeatureType]:
        """Create a new instance from sample, Roi and Feature data.

        :param sample: the sample associated with the data
        :param rois: a dictionary that maps snapshot ids to a list of ROIs in the snapshot
        :param features: a dictionary that maps snapshot ids to a list of features in the snapshot
        :param snapshots: the list of snapshots to create
        :param state_list: a mapping from snapshot id to sample data state

        """
        data = cls(sample, roi_type, feature_type)
        for snapshot_id in snapshots:
            snapshot_roi = rois.get(snapshot_id)
            if snapshot_roi is None:
                raise ValueError(f"rois dictionary must contain a ROI list for snapshot {snapshot_id}.")
            snapshot_features = features.get(snapshot_id)

            if snapshot_features is None:
                raise ValueError(f"rois dictionary must contain a ROI list for snapshot {snapshot_id}.")

            snapshot_state = states.get(snapshot_id)
            if snapshot_state is None:
                raise ValueError(f"states dictionary must contain a state for snapshot {snapshot_id}.")

            snapshot = OnMemorySampleStorageSnapshot(sample, snapshot_id, snapshot_state, roi_type, feature_type)
            snapshot.add_rois(*snapshot_roi)
            snapshot.add_features(*snapshot_features)
            data._add_snapshot(snapshot)
        return data


class OnMemorySampleStorageSnapshot(Generic[RoiType, FeatureType]):
    """Stores data state during a sample processing pipeline."""

    def __init__(
        self,
        sample: Sample,
        snapshot_id: str,
        status: SampleProcessStatus,
        roi_type: type[RoiType],
        feature_type: type[FeatureType],
    ):
        self._roi_type = roi_type
        self._feature_type = feature_type
        self.sample = sample
        self.id = snapshot_id
        self._features: dict[UUID, FeatureType] = dict()
        self._rois: dict[UUID, RoiType] = dict()
        self._roi_to_features: dict[UUID, set[UUID]] = dict()
        self.status = status

    def copy(self, snapshot_id: str, set_new_ids: bool = False) -> OnMemorySampleStorageSnapshot[RoiType, FeatureType]:
        """Create a copy of the snapshot."""
        status = self.status.model_copy()
        copy = OnMemorySampleStorageSnapshot(self.sample, snapshot_id, status, self._roi_type, self._feature_type)

        rois_copy = list()
        roi_map = dict()
        for roi in self.list_rois():
            roi_copy = roi.model_copy(deep=True)
            rois_copy.append(roi_copy)
            if set_new_ids:
                roi_copy.id = create_id()
            roi_map[roi.id] = roi_copy

        features_copy = list()
        for ft in self.list_features():
            ft_copy = ft.model_copy(deep=True, update={"roi": roi_map[ft.roi.id]})
            features_copy.append(ft_copy)
            if set_new_ids:
                ft_copy.id = create_id()
                ft_copy.annotation.id = ft_copy.id  # type: ignore

        copy.add_rois(*rois_copy)
        copy.add_features(*features_copy)
        return copy

    def add_features(self, *features: FeatureType) -> None:
        """Add features to the snapshot."""
        new_features = dict()
        for ft in features:
            if not self.has_roi(ft.roi.id):
                msg = f"Cannot add feature {ft.id} because its parent ROI {ft.roi.id} was not found in the data."
                raise exceptions.RoiNotFound(msg)

            if self.has_feature(ft.id):
                msg = f"A feature with id={ft.id} already exists in {self.sample.id}/{self.id}"
                raise exceptions.RepeatedIdError(msg)

            new_features[ft.id] = ft
        self._features.update(new_features)

        for ft_id, ft in new_features.items():
            self._roi_to_features[ft.roi.id].add(ft_id)

    def add_rois(self, *rois: RoiType) -> None:
        """Add Rois to the snapshot."""
        new_rois = dict()
        new_roi_to_features = dict()
        for roi in rois:
            if self.has_roi(roi.id):
                msg = f"ROI with id {roi.id} already stored in {self.sample.id}/{self.id}."
                raise exceptions.RepeatedIdError(msg)
            new_rois[roi.id] = roi
            new_roi_to_features[roi.id] = set()
        self._rois.update(new_rois)
        self._roi_to_features.update(new_roi_to_features)

    def delete_features(self, *feature_ids: UUID) -> None:
        """Delete features from the snapshot."""
        for id_ in feature_ids:
            if not self.has_feature(id_):
                continue
            ft = self._features.pop(id_)
            parent_roi_features = self._roi_to_features[ft.roi.id]
            parent_roi_features.remove(id_)

    def delete_rois(self, *roi_ids: UUID) -> None:
        """Delete ROIs and their associated features."""
        for id_ in roi_ids:
            if not self.has_roi(id_):
                continue

            del self._rois[id_]

            features = self._roi_to_features.pop(id_)
            for child_id in features:
                del self._features[child_id]

    def get_feature(self, feature_id: UUID) -> FeatureType:
        """Retrieve a feature by id."""
        feature = self._features.get(feature_id)
        if feature is None:
            msg = f"Feature {feature_id} not found  in {self.sample.id}/{self.id}."
            raise exceptions.FeatureNotFound(msg)
        return feature

    def get_roi(self, roi_id: UUID) -> RoiType:
        """Retrieve a ROI by id."""
        roi = self._rois.get(roi_id)
        if roi is None:
            msg = f"ROI {roi_id} not found in {self.sample.id}/{self.id}."
            raise exceptions.RoiNotFound(msg)
        return roi

    def get_n_features(self) -> int:
        """Get the total number of features in the snapshot."""
        return len(self._features)

    def get_n_rois(self) -> int:
        """Get the total number of ROIs in the snapshot."""
        return len(self._rois)

    def has_feature(self, feature_id: UUID) -> bool:
        """CHeck the existence of a feature using its id."""
        return feature_id in self._features

    def has_roi(self, roi_id: UUID) -> bool:
        """Check the existence of a ROI."""
        return roi_id in self._rois

    def list_rois(self) -> list[RoiType]:
        """List all ROIs in the sample storage."""
        return list(self._rois.values())

    def list_features(self, roi_id: UUID | None = None) -> list[FeatureType]:
        """List all features in the snapshot."""
        if roi_id is None:
            return list(self._features.values())

        if not self.has_roi(roi_id):
            msg = f"Roi {roi_id} not found in {self.sample.id}/{self.id}."
            raise exceptions.RoiNotFound(msg)

        return [self.get_feature(x) for x in self._roi_to_features[roi_id]]

    @classmethod
    def from_sample_storage(
        cls, sample_storage: SampleStorage[RoiType, FeatureType]
    ) -> OnMemorySampleStorageSnapshot[RoiType, FeatureType]:
        """Create a snapshot from sample storage current status."""
        sample = sample_storage.get_sample()
        status = sample_storage.get_status()
        snapshot_id = sample_storage.get_snapshot_id()
        snapshot = cls(sample, snapshot_id, status, sample_storage.roi_type, sample_storage.feature_type)
        snapshot.add_rois(*sample_storage.list_rois())
        snapshot.add_features(*sample_storage.list_features())
        return snapshot.copy(snapshot_id, set_new_ids=False)


class OnMemoryAssayStorageSnapshot:
    """Store independent copies of assay descriptors and missing values."""

    def __init__(self, snapshot_id: str, status: AssayProcessStatus):
        self.id = snapshot_id
        self.status = status
        self._table = FeatureTable()

    def add_descriptors(self, descriptors: list[dict[str, float]], annotations: list[Annotation]):
        """Add annotations and descriptors from a sample."""
        self._table.add_descriptors(descriptors, annotations)

    def add_feature_groups(self, *feature_groups: FeatureGroup) -> None:
        """Add feature groups to the snapshot."""
        self._table.add_feature_groups(*feature_groups)

    def fetch_feature_groups(self) -> list[FeatureGroup]:
        """Fetch feature groups from the snapshot."""
        return self._table.fetch_feature_groups()

    def has_feature_group(self, group: int) -> bool:
        """Check if a group with the provided id is stored in the assay."""
        return self._table.has_feature_group(group)

    def fetch_annotations(self, sample_id: str | None = None, copy: bool = False) -> list[Annotation]:
        """Create a list feature annotations.

        :param sample_id: If provided, only include annotations from this sample
        """
        return self._table.fetch_annotations(sample_id=sample_id, copy=copy)

    def fetch_descriptors(
        self, descriptors: Iterable[str] | None = None, sample_id: str | None = None, copy: bool = False
    ) -> dict[str, list[float]]:
        """Fetch descriptors from the snapshot."""
        return self._table.fetch_descriptors(descriptors=descriptors, sample_id=sample_id, copy=copy)

    def fetch_fill_values(self, copy: bool = False) -> dict[str, dict[int, float]]:
        """Fetch snapshot fill values."""
        return self._table.fetch_fill_values(copy=copy)

    def add_fill_values(self, fill_values: dict[str, dict[int, float]]) -> None:
        """Add missing values to the snapshot."""
        self._table.add_fill_values(fill_values)

    def get_sample_id(self, feature_id: UUID) -> str:
        """Retrieve the sample id of a feature."""
        return self._table.get_sample_id(feature_id)

    def get_ids_by_group(self, group: int) -> list[UUID]:
        """Retrieve all feature ids associated with a feature group."""
        return self._table.get_ids_by_group(group)

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a feature is in the snapshot."""
        return self._table.has_feature(feature_id)

    def patch_annotation(self, *patches: AnnotationPatch) -> None:
        """Apply patches to annotations."""
        self._table.patch_annotation(*patches)

    def list_feature_groups(self) -> list[int]:
        """List all feature groups stored in the assay."""
        return self._table.list_feature_groups()

    def patch_descriptors(self, *patches: DescriptorPatch) -> None:
        """Apply patches to descriptors."""
        self._table.patch_descriptors(*patches)

    def copy_snapshot(self, copy_id: str) -> OnMemoryAssayStorageSnapshot:
        """Create a snapshot copy."""
        res = OnMemoryAssayStorageSnapshot(copy_id, self.status.model_copy(deep=True))
        descriptor_list = self.fetch_descriptors(copy=True)
        annotations = self.fetch_annotations(copy=True)
        fill_values = self.fetch_fill_values(copy=True)
        descriptors_names = list(descriptor_list)
        descriptors = [dict(zip(descriptors_names, values)) for values in zip(*descriptor_list.values())]
        res.add_descriptors(descriptors, annotations)
        res.add_fill_values(fill_values)
        return res



import numpy as np
import pydantic
import pytest

from tidyms2.core.utils.numpy import FloatArray, IntArray, find_closest


class NumpyFloatModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)
    arr: FloatArray


class NumpyIntModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)
    arr: IntArray


class TestSerializeFloatArray:
    def test_empty_array(self):
        expected = np.array([], dtype=float)
        model = NumpyFloatModel(arr=expected)
        actual = NumpyFloatModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)

    def test_1D_array(self):
        expected = np.random.normal(size=100)
        model = NumpyFloatModel(arr=expected)
        actual = NumpyFloatModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)

    def test_2D_array(self):
        expected = np.random.normal(size=(100, 20))
        model = NumpyFloatModel(arr=expected)
        actual = NumpyFloatModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)


class TestSerializeIntArray:
    def test_empty_array(self):
        expected = np.array([], dtype=int)
        model = NumpyIntModel(arr=expected)
        actual = NumpyIntModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)

    def test_1D_array(self):
        expected = np.random.randint(low=0, high=100, size=100)
        model = NumpyIntModel(arr=expected)
        actual = NumpyIntModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)

    def test_2D_array(self):
        expected = np.random.randint(low=0, high=100, size=(100, 25))
        model = NumpyIntModel(arr=expected)
        actual = NumpyIntModel(**model.model_dump()).arr
        assert np.array_equal(actual, expected)


class TestFindClosest:
    def test_find_closest_left_border(self):
        x = np.arange(10)
        y = -1
        ind = find_closest(x, y)
        assert ind == 0

    def test_find_closest_right_border(self):
        x = np.arange(10)
        y = 10
        ind = find_closest(x, y)
        assert ind == (x.size - 1)

    def test_find_closest_middle(self):
        x = np.arange(10)
        y = 4.6
        ind = find_closest(x, y)
        assert ind == 5

    def test_find_closest_empty_x(self):
        x = np.array([])
        y = 10
        with pytest.raises(ValueError):
            find_closest(x, y)

    def test_find_closest_empty_y(self):
        x = np.arange(10)
        y = np.array([])
        res = find_closest(x, y)
        assert res.size == 0

    def test_find_closest_multiple_values(self):
        x = np.arange(100)
        y = np.array([-10, 4.6, 67.1, 101])
        ind = np.array([0, 5, 67, 99], dtype=int)
        result = find_closest(x, y)
        assert np.array_equal(result, ind)

    def test_find_closest_unsorted_single_value(self):
        n = 100
        x = np.random.normal(size=n)
        # select three random points
        random_index = np.random.choice(n)
        xq = x[random_index]
        closest_index = find_closest(x, xq, is_sorted=False)
        assert np.equal(random_index, closest_index).all()

    def test_find_closest_unsorted_multiple_values(self):
        n = 100
        x = np.random.normal(size=n)
        # select three random points
        random_index = np.random.choice(n, 3)
        xq = x[random_index]
        closest_index = find_closest(x, xq, is_sorted=False)
        assert np.equal(random_index, closest_index).all()



"""Implementation of the consensus annotation algorithm."""

from __future__ import annotations

from typing import Generator

import pydantic

from ..core.models import Annotation, GroupAnnotation


class MissingNodeError(ValueError):
    """Exception raised when a missing node group if requested to an annotation graph."""


class AnnotationGraph:
    """Store consensus annotation nodes and edges.

    A node represents a feature group, which stores charge status and isotopologue
    indices annotated on each assay sample.

    An edge represent a relation between an :term:`MMI` and another isotopologue
    in the group. It also stores the number of samples where the annotation was
    detected.

    """

    def __init__(self):
        self.nodes: dict[int, AnnotationNode] = dict()
        self.edges: dict[int, dict[int, int]] = dict()

    def __iter__(self) -> Generator[AnnotationNode, None, None]:
        for node in self.nodes.values():
            yield node

    def add_annotation(self, annotation: Annotation) -> None:
        """Add a new node or update an existing node using a feature annotation."""
        if annotation.group in self.nodes:
            node = self.nodes[annotation.group]
        else:
            node = AnnotationNode(group=annotation.group)
            self.nodes[annotation.group] = node

        node.charge_votes.setdefault(annotation.charge, 0)
        node.charge_votes[annotation.charge] += 1
        node.index_votes.setdefault(annotation.isotopologue_index, 0)
        node.index_votes[annotation.isotopologue_index] += 1

    def add_edge(self, group1: int, group2: int) -> None:
        """Create a new edge or update the edge count."""
        node1 = self.get_node(group1)
        node2 = self.get_node(group2)
        node1.neighbors.add(group2)
        node2.neighbors.add(group1)

        group1_edges = self.edges.setdefault(group1, dict())
        group1_edges.setdefault(group2, 0)
        group1_edges[group2] += 1

        group2_edges = self.edges.setdefault(group2, dict())
        group2_edges.setdefault(group1, 0)
        group2_edges[group1] += 1

    def get_edges(self, group: int) -> list[AnnotationEdge]:
        """Get node edges."""
        edges = list()
        node = self.get_node(group)
        node_edges_dict = self.edges.get(group, dict())
        for neighbor_group, edge_votes in node_edges_dict.items():
            neighbor = self.get_node(neighbor_group)
            edge = AnnotationEdge(groups=(node.group, neighbor.group), index=neighbor.index, votes=edge_votes)
            edges.append(edge)
        return edges

    def get_neighbors(self, group: int) -> list[AnnotationNode]:
        """Retrieve all neighbor nodes."""
        return [self.get_node(x) for x in self.get_node(group).neighbors]

    def get_node(self, group: int) -> AnnotationNode:
        """Fetch a node."""
        try:
            return self.nodes[group]
        except KeyError as e:
            msg = f"Node group {group} not found."
            raise MissingNodeError(msg) from e

    def list_node_groups(self) -> list[int]:
        """Retrieve the unique node groups in the graph."""
        return list(self.nodes)

    def remove_isolated(self) -> None:
        """Remove nodes with no neighbors."""
        for group in self.list_node_groups():
            node = self.get_node(group)
            if not node.neighbors:
                self.remove_node(node.group)

    def remove_node(self, group: int) -> None:
        """Remove a node if it exists."""
        edges = self.get_edges(group)
        for edge in edges:
            self.remove_edge(*edge.groups)
        self.nodes.pop(group)

    def remove_edge(self, group1: int, group2: int) -> None:
        """Remove an edge if it exists."""
        node1 = self.get_node(group1)
        node2 = self.get_node(group2)

        if group2 in node1.neighbors and group1 in node2.neighbors:
            self.edges[group1].pop(group2)
            self.edges[group2].pop(group1)

            node1.neighbors.remove(group2)
            node2.neighbors.remove(group1)
        else:
            msg = f"There is no edge between annotation node {group1} and {group2}."
            raise ValueError(msg)

    def count_votes(self) -> None:
        """Update node index and charge based on total counts."""
        for node in self:
            node.charge = max(node.charge_votes, key=lambda x: node.charge_votes[x])
            node.index = max(node.index_votes, key=lambda x: node.index_votes[x])


class AnnotationNode(pydantic.BaseModel):
    """Store a node with annotations."""

    group: int
    """The node id, equal to the feature group."""

    charge_votes: dict[int, int] = dict()
    """Maps charge state to number of feature occurrences."""

    index_votes: dict[int, int] = dict()
    """Maps isotopologue index to number of feature occurrences."""

    charge: int = -1
    """The most voted charge state"""

    index: int = -1
    """The most voted isotopologue index"""

    neighbors: set[int] = set()
    """The node neighbors"""


class AnnotationEdge(pydantic.BaseModel):
    """Store a node with annotations."""

    groups: tuple[int, int]
    """The pair of groups connected by the edge."""

    index: int
    """The isotopologue index that connects with the :term:`MMI`."""

    votes: int
    """The edge votes."""

    approval: float = pydantic.Field(default=0.0, ge=0.0, le=1.0)
    """The approval metric for the annotation edge."""


def create_consensus_annotation(annotations: list[Annotation]) -> dict[int, GroupAnnotation]:
    """Create the annotations for all feature groups."""
    graph = create_annotation_graph(annotations)
    consensus = dict()

    # provide a common isotopologue group id
    all_mmi = {x.group for x in graph if x.index == 0}
    isotopologue_group_id = 0
    while all_mmi:
        mmi = graph.get_node(all_mmi.pop())
        isotopologues_group = graph.get_neighbors(mmi.group)
        isotopologues_group.append(mmi)
        isotopologues_group = trim_isotopologues(isotopologues_group)
        if not isotopologues_group:
            continue
        for isotopologues in isotopologues_group:
            consensus[isotopologues.group] = GroupAnnotation(
                label=isotopologues.group,
                isotopologue_group=isotopologue_group_id,
                isotopologue_index=isotopologues.index,
                charge=isotopologues.charge,
            )
        isotopologue_group_id += 1
    return consensus


def create_annotation_graph(annotations: list[Annotation]) -> AnnotationGraph:
    """Create an annotation graph and add all annotations nodes and edges."""
    graph = AnnotationGraph()
    for sample_annotations in group_annotations_by_sample(annotations).values():
        add_sample_annotations(graph, sample_annotations)

    graph.count_votes()

    # first remove edges that connect multiple MMIs or edges that connect MMI with
    # nodes with a different charge state
    for node in graph:
        if node.index == 0:
            remove_invalid_mmi_neighbors(graph, node)

    # then for non-MMI nodes keep only the most voted edge
    for node in graph:
        if node.index != 0:
            solve_non_mmi_conflict(graph, node)

    # finally remove repeated edges connecting MMI nodes to nodes with repeated
    # isotopologue indices
    for node in graph:
        if node.index == 0:
            solve_mmi_conflict(graph, node)

    graph.remove_isolated()

    return graph


def add_sample_annotations(graph: AnnotationGraph, annotations: list[Annotation]):
    """Add annotations from a single sample."""
    isotopologues_dict: dict[int, list[int]] = dict()
    mmi_dict: dict[int, int] = dict()

    for ann in annotations:
        if ann.group == -1 or ann.isotopologue_label == -1:
            continue
        graph.add_annotation(ann)
        if ann.isotopologue_index == 0:
            mmi_dict[ann.isotopologue_label] = ann.group
        else:
            isotopologues = isotopologues_dict.setdefault(ann.isotopologue_label, list())
            isotopologues.append(ann.group)

    for mmi_label, mmi_group in mmi_dict.items():
        isotopologues = isotopologues_dict.get(mmi_label)
        if isotopologues is None:
            continue

        for i in isotopologues:
            graph.add_edge(mmi_group, i)


def group_annotations_by_sample(annotations: list[Annotation]) -> dict[str, list[Annotation]]:
    """Group annotations by sample."""
    sample_to_annotations = dict()
    for ann in annotations:
        sample_annotations = sample_to_annotations.setdefault(ann.sample_id, list())
        sample_annotations.append(ann)
    return sample_to_annotations


def remove_invalid_mmi_neighbors(graph: AnnotationGraph, mmi: AnnotationNode):
    """Remove MMI neighbors with different charge state or equal indices."""
    rm_list = list()
    for neighbor_id in mmi.neighbors:
        neighbor = graph.get_node(neighbor_id)
        has_different_charge = neighbor.charge != mmi.charge
        has_equal_index = neighbor.index == mmi.index
        if has_different_charge or has_equal_index:
            rm_list.append(neighbor.group)

    for rm_group in rm_list:
        graph.remove_edge(mmi.group, rm_group)


def solve_mmi_conflict(graph: AnnotationGraph, mmi: AnnotationNode):
    """Remove neighbors with repeated index.

    Only the most voted is kept.
    """
    edges = graph.get_edges(mmi.group)
    index_to_edge: dict[int, list[AnnotationEdge]] = dict()
    for edge in edges:
        i_edges = index_to_edge.setdefault(edge.index, list())
        i_edges.append(edge)

    for i_edges in index_to_edge.values():
        most_voted_edge = max(i_edges, key=lambda x: x.votes)
        for edge in i_edges:
            if edge != most_voted_edge:
                graph.remove_edge(*edge.groups)


def solve_non_mmi_conflict(graph: AnnotationGraph, node: AnnotationNode) -> None:
    """Keep only the most voted neighbor for non-MMI nodes."""
    edges = graph.get_edges(node.group)

    if len(edges) > 1:
        most_voted_edge = max(edges, key=lambda x: x.votes)
        for edge in edges:
            if edge != most_voted_edge:
                graph.remove_edge(*edge.groups)


def compute_approval(edges: list[AnnotationEdge]) -> None:
    """Compute an edge approval metric."""
    # TODO: complete
    total = sum(x.votes for x in edges)
    for edge in edges:
        edge.approval = edge.votes / total


def trim_isotopologues(isotopologues: list[AnnotationNode]) -> list[AnnotationNode]:
    """Trim invalid feature annotations.

    Annotations are trimmed in the following cases:

    - isotopologue index 0 is not part of the index list in each isotopologue group
    - there are missing indices, i.e., 0, 1, 3, 4, ...

    """
    trimmed = list()
    for k, node in enumerate(sorted(isotopologues, key=lambda x: x.index)):
        if node.index != k:
            break
        trimmed.append(node)
    return trimmed



# noqa: E402
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from sphinx_pyproject import SphinxConfig

config = SphinxConfig("../pyproject.toml", globalns=globals())

project = config.name
copyright = "2024, Gabriel Riquelme"
author = config.author
release = config.version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
import os
import sys

sys.path.append(os.path.abspath("../src"))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.autodoc_pydantic",
    "matplotlib.sphinxext.plot_directive",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_theme_options = {
  "show_nav_level": 2
}

# autodoc-pydantic
# ----------------
autodoc_pydantic_model_show_json = False
autodoc_pydantic_settings_show_json = False
autodoc_pydantic_model_show_field_summary = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_show_validator_members = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_model_undoc_members = False
autodoc_pydantic_field_list_validators = False
autodoc_pydantic_field_show_constraints = False


"""Utilities to build LC-MS samples."""

import csv
import pathlib
from typing import Any

import yaml

from ..core.exceptions import SampleNotFound
from ..core.models import Sample, SampleMetadata


def create_csv_sample_template(
    pattern: str,
    template_path: pathlib.Path | str,
    loc: pathlib.Path | None = None,
    **kwargs,
) -> None:
    """Create a csv file template will all files in the data location that matches the provided pattern.

    :param pattern: A glob expression to select files. For example, passing ``"*.mzML"`` will select
        all files with mzML extension in the provided location.
    :param template_path: path where the template file
    :param loc: The directory where the glob expression will be evaluated. If not provided, it will
        be set to the current working directory.
    :param kwargs: default values passed to sample parameters, e.g. `start_time`, `ms_level`, `reader`, etc...
    :raises SampleNotFound: If no samples are found at the provided location.

    """
    loc = loc or pathlib.Path.cwd()

    samples: list[dict[str, Any]] = list()

    for order, file in enumerate(sorted(loc.glob(pattern), key=lambda x: x.name), start=1):
        sample = Sample(id=file.stem, path=file, **kwargs)
        sample.meta.order = order

        d = sample.model_dump()
        meta_d = d.pop("meta")
        d.update(meta_d)
        samples.append(d)

    if not samples:
        raise SampleNotFound(f"No samples at {loc} match the pattern `{pattern}`.")

    if isinstance(template_path, str):
        template_path = pathlib.Path(template_path)

    with template_path.open("wt") as f_out:
        headers = list(samples[0])

        writer = csv.DictWriter(f_out, headers)
        writer.writeheader()
        writer.writerows(samples)


def create_yaml_sample_template(
    pattern: str,
    template_path: pathlib.Path | str,
    loc: pathlib.Path | None = None,
    **kwargs,
) -> None:
    """Create a csv file template will all files in the data location that matches the provided pattern.

    :param pattern: A glob expression to select files. For example, passing ``"*.mzML"`` will select
        all files with mzML extension in the provided location.
    :param template_path: path where the template file
    :param loc: The directory where the glob expression will be evaluated. If not provided, it will
        be set to the current working directory.
    :param kwargs: default values passed to sample parameters, e.g. `start_time`, `ms_level`, `reader`, etc...
    :raises SampleNotFound: If no samples are found at the provided location.

    """
    loc = loc or pathlib.Path.cwd()

    samples: list[dict[str, Any]] = list()

    for order, file in enumerate(sorted(loc.glob(pattern), key=lambda x: x.name), start=1):
        sample = Sample(id=file.stem, path=file, **kwargs)
        sample.meta.order = order
        samples.append(sample.model_dump(mode="json"))

    if not samples:
        raise SampleNotFound(f"No samples at {loc} match the pattern `{pattern}`.")

    if isinstance(template_path, str):
        template_path = pathlib.Path(template_path)

    with template_path.open("wt") as file:
        yaml.dump(samples, file, yaml.Dumper)


def read_samples_from_yaml_template(template_path: pathlib.Path | str) -> list[Sample]:
    """Create samples using a YAML template file created with :py:func:`create_yaml_sample_template`.

    :param template_path: the path to the template file.

    """
    if isinstance(template_path, str):
        template_path = pathlib.Path(template_path)

    with template_path.open("rt") as file:
        samples = yaml.load(file, yaml.Loader)

    assert isinstance(samples, list), "Expected YAML file to contain a list of sample models."

    return [Sample(**d) for d in samples]


def read_samples_from_csv_template(template_path: pathlib.Path | str) -> list[Sample]:
    """Create samples using a CSV template file created with :py:func:`create_csv_sample_template`.

    :param template_path: the path to the template file.

    """
    if isinstance(template_path, str):
        template_path = pathlib.Path(template_path)

    with template_path.open("rt") as fin:
        dialect = csv.Sniffer().sniff(fin.read(2048))
        fin.seek(0)
        reader = csv.DictReader(fin, dialect=dialect)

        sample_fields = list(Sample.model_fields)
        samples: list[Sample] = list()
        for row in reader:
            row = {k: v for k, v in row.items() if v}  # ignore empty records in csv
            meta = SampleMetadata(**{k: v for k, v in row.items() if k not in sample_fields})  # type: ignore
            sample = Sample(meta=meta, **{k: v for k, v in row.items() if k in sample_fields})  # type: ignore
            samples.append(sample)
    return samples



import pytest

from tidyms2.core.enums import SampleType
from tidyms2.simulation.utils import create_sample_list


def test_simulate_empty_sample_list():
    simulated = create_sample_list(list())
    assert not simulated


def test_simulate_no_groups_groups_are_equal_to_sample_type():
    sample_types = [SampleType.TECHNICAL_QC, SampleType.SAMPLE]
    simulated = create_sample_list(sample_types)
    assert len(simulated) == len(sample_types)
    for sample in simulated:
        assert sample.meta.type == sample.meta.group


def test_simulate_with_start_order():
    sample_types = [SampleType.TECHNICAL_QC, SampleType.SAMPLE]
    start_at = 5
    simulated = create_sample_list(sample_types, start_order=start_at)
    assert len(simulated) == len(sample_types)
    for order, sample in enumerate(simulated, start=start_at):
        assert sample.meta.type == sample.meta.group
        assert order == sample.meta.order


def test_simulate_with_type_to_groups_missing_group_raises_error():
    sample_types = [SampleType.TECHNICAL_QC, SampleType.SAMPLE]
    type_to_groups = {SampleType.SAMPLE: ["group1"]}
    with pytest.raises(ValueError):
        create_sample_list(sample_types, type_to_groups=type_to_groups)


def test_simulate_with_type_to_groups():
    sample_types = [SampleType.TECHNICAL_QC, SampleType.SAMPLE]
    type_to_groups = {SampleType.SAMPLE: ["group1"], SampleType.TECHNICAL_QC: ["QC"]}
    simulated = create_sample_list(sample_types, type_to_groups=type_to_groups)
    for sample in simulated:
        # FIXME: fix when SampleMeta.type is a SampleType
        assert sample.meta.group in type_to_groups[sample.meta.type]  # type: ignore


def test_simulate_multiple_batches():
    sample_types = [SampleType.TECHNICAL_QC, SampleType.SAMPLE]
    n_batches = 5
    simulated = create_sample_list(sample_types, n_batches=n_batches)
    for order, sample in enumerate(simulated):
        assert sample.meta.order == order
        assert sample.meta.batch == order // len(sample_types)



"""Utilities to simulate LC-MS data."""

from __future__ import annotations

import pathlib
import random

import numpy
import pydantic

from ..core.matrix import DataMatrix
from ..core.models import Chromatogram, FeatureGroup, GroupAnnotation, MSSpectrum, Sample, SampleMetadata
from ..core.utils.numpy import FloatArray, FloatArray1D
from ..io.reader import reader_registry
from .base import BaseChemicalSpeciesSpec, DataAcquisitionSpec


@reader_registry.register()
class SimulatedLCMSDataReader:
    """Read simulated LC-MS data files."""

    def __init__(self, src: pathlib.Path | Sample) -> None:
        if isinstance(src, pathlib.Path) or not hasattr(src.meta, "_simulated"):
            msg = "Simulated LC-MS sample only work with sample models created with the simulated sample factory."
            raise ValueError(msg)
        self._simulated_sample_spec = src.meta._simulated  # type: ignore
        self._spectrum_factory = MSSpectrumFactory(self._simulated_sample_spec)

    def get_chromatogram(self, index: int) -> Chromatogram:
        """Retrieve a chromatogram from file."""
        raise NotImplementedError

    def get_spectrum(self, index: int) -> MSSpectrum:
        """Retrieve a spectrum from file."""
        return self._spectrum_factory.create(index)

    def get_n_chromatograms(self) -> int:
        """Retrieve the total number of chromatograms."""
        raise NotImplementedError

    def get_n_spectra(self) -> int:
        """Retrieve the total number of spectra."""
        return self._simulated_sample_spec.config.n_scans


class SimulatedLCMSSampleFactory(pydantic.BaseModel):
    """Utility that creates simulated data samples.

    Refer to the :ref:`simulation-guide` guide for a tutorial on how to use this class.

    """

    data_acquisition: DataAcquisitionSpec = DataAcquisitionSpec()
    """The sample configuration used to simulate data."""

    adducts: list[SimulatedLCMSAdductSpec] = list()
    """the list of adducts to include in the simulated sample."""

    def __call__(self, id: str, group: str | None = None, order: int = 0, **kwargs) -> Sample:
        """Create a :py:class:`~tidyms2.core.models.Sample` with simulated data specs.

        Refer to the :ref:`simulation-guide` guide for a tutorial on how to use this function.

        :param id: the id for the sample
        :param kwargs: extra sample information passed to the :py:class:`tidyms2.Sample` constructor.

        """
        if "path" not in kwargs:
            kwargs["path"] = pathlib.Path(".")

        reader = SimulatedLCMSDataReader.__name__
        meta = SampleMetadata(group=group or "", order=order)
        meta._simulated = self.create_simulated_sample_spec(group, order)  # type: ignore
        return Sample(id=id, reader=reader, meta=meta, **kwargs)

    def create_features(self, group: str | None = None, order: int = 0) -> list[SimulatedLCMSFeature]:
        """Create the list of features in a sample using the spec."""
        features: list[SimulatedLCMSFeature] = list()
        for adduct in self.adducts:
            features.extend(adduct.create_simulated_features(group, order))
        return sorted(features, key=lambda x: x.mz)

    def create_simulated_sample_spec(self, group: str | None = None, order: int = 0) -> SimulatedLCMSSample:
        """Create a simulated sample specification."""
        features = self.create_features(group, order)
        return SimulatedLCMSSample(config=self.data_acquisition, features=features)


class SimulatedLCMSSample(pydantic.BaseModel):
    """Create simulated LC-MS data files."""

    config: DataAcquisitionSpec
    """The sample configuration used to simulate data."""

    features: list[SimulatedLCMSFeature] = list()
    """the list of features in the sample."""

    def make_grid(self) -> FloatArray1D:
        """Create a grid from features m/z values."""
        if self.config.grid is None:
            grid = numpy.array(sorted([x.mz for x in self.features]))
        else:
            grid = self.config.grid.create()
        return grid


class RtSpec(pydantic.BaseModel):
    """Define the retention time value in samples."""

    mean: pydantic.PositiveFloat = 10.0
    """The mean retention time across samples"""

    std: pydantic.NonNegativeFloat = 0.0
    """The retention time standard deviation across samples"""

    width: pydantic.PositiveFloat = 3.0
    """The peak width of the chromatogram"""

    def compute_sample_rt(self) -> float:
        """Compute an observation of the retention time."""
        noise = random.gauss(sigma=self.std)
        return max(0.0, self.mean + noise)


class SimulatedLCMSAdductSpec(BaseChemicalSpeciesSpec):
    """Define a set of isotopologue feature created from an adduct."""

    rt: RtSpec = RtSpec()
    """The adduct retention time specification"""

    def create_simulated_features(self, group: str | None = None, order: int = 0) -> list[SimulatedLCMSFeature]:
        """Create a list of simulated features used by a simulated sample."""
        rt = self.rt.compute_sample_rt()
        features = list()
        for mzk, ik in zip(self.get_mz(), self.compute_intensity(group, order)):
            ft = SimulatedLCMSFeature(mz=mzk, rt=rt, int=ik, width=self.rt.width)
            features.append(ft)
        return features


class SimulatedLCMSFeature(pydantic.BaseModel):
    """Store a simulated LC-MS peak information."""

    mz: pydantic.PositiveFloat
    """The feature m/z."""

    rt: pydantic.PositiveFloat
    """The feature retention time."""

    int: pydantic.NonNegativeFloat
    """the feature intensity."""

    width: pydantic.PositiveFloat
    """The peak width in the time domain"""


class MSSpectrumFactory:
    """Store m/z grid data."""

    def __init__(self, sample: SimulatedLCMSSample) -> None:
        self.sample = sample
        self.grid = sample.make_grid()
        self._is_centroid = sample.config.grid is not None
        scan = sample.config.n_scans
        self._random_seeds = numpy.random.choice(scan * 10, scan)

    def create(self, scan: int) -> MSSpectrum:
        """Create a ms spectrum instance."""
        assert scan < self.sample.config.n_scans, "`scan` must be lower than the sample `n_scans` parameter."

        mz = self._compute_mz(scan)
        sp = self._compute_intensity(mz, scan)

        if self.sample.config.min_int is not None:
            mask = sp >= self.sample.config.min_int
            mz = mz[mask]
            sp = sp[mask]

        time = self.sample.config.time_resolution * scan

        return MSSpectrum(
            index=scan, mz=mz, int=sp, ms_level=self.sample.config.ms_level, centroid=self._is_centroid, time=time
        )

    def _compute_mz(self, scan: int) -> FloatArray1D:
        # use the same random seed for a given scan for reproducibility
        seed = self._random_seeds[scan]
        numpy.random.seed(seed)

        # add random noise to the m/z grid
        noise_level = self.sample.config.mz_std

        if noise_level > 0.0:
            noise = numpy.random.normal(size=self.grid.size, scale=noise_level)
            mz = self.grid + noise
        else:
            mz = self.grid.copy()

        return mz

    def _compute_intensity(self, mz: FloatArray1D, scan: int):
        time = self.sample.config.time_resolution * scan
        intensity = numpy.zeros_like(mz)
        mz_width = self.sample.config.mz_width
        for ft in self.sample.features:
            amp = ft.int * numpy.power(numpy.e, -0.5 * ((time - ft.rt) / ft.width) ** 2)

            # this step allows computing values both in profile and centroid mode
            intensity += amp * numpy.power(numpy.e, -0.5 * ((mz - ft.mz) / mz_width) ** 2)

        if self.sample.config.int_std > 0.0:
            intensity += numpy.random.normal(size=intensity.size, scale=self.sample.config.int_std)
            intensity[intensity < 0] = 0.0

        return intensity


def simulate_data_matrix(specs: list[SimulatedLCMSAdductSpec], samples: list[Sample]) -> DataMatrix:
    """Create a data matrix using LC-MS simulation specs.

    Refer to the :ref:`matrix-simulation-guide` guide for a tutorial on how to use this function.

    :param specs: a list of adduct specifications to simulate features.
    :param samples: the list of samples to include in the matrix.
    :return: the simulated data matrix

    """
    if not samples:
        raise ValueError("At least one sample is required to simulate a data matrix.")

    # sorting samples by order is required to compute intra-batch effects
    samples = sorted(samples, key=lambda x: x.meta.order)

    X = _compute_matrix_values(specs, samples)
    feature_groups = _compute_feature_groups(specs)

    return DataMatrix(samples, feature_groups, X)


def _compute_matrix_values(specs: list[SimulatedLCMSAdductSpec], samples: list[Sample]) -> FloatArray:
    n_features = sum(x.n_isotopologues for x in specs)
    n_samples = len(samples)
    X = numpy.zeros(shape=(n_samples, n_features), dtype=float)

    first_sample = samples[0]
    previous_sample_batch = first_sample.meta.batch
    batch_start_at = first_sample.meta.order

    for k, sample in enumerate(samples):
        row = list()

        if previous_sample_batch != sample.meta.batch:
            batch_start_at = sample.meta.order
            previous_sample_batch = sample.meta.batch
        intra_batch_order = sample.meta.order - batch_start_at

        for spec in specs:
            row.extend(
                spec.compute_intensity(
                    group=sample.meta.group,
                    order=intra_batch_order,
                    batch=sample.meta.batch,
                )
            )
        X[k] = row
    return X


def _compute_feature_groups(specs: list[SimulatedLCMSAdductSpec]) -> list[FeatureGroup]:
    feature_groups = list()
    group_counter = 0

    for isotopologue_group, spec in enumerate(specs):
        for isotopologue_index, ft in enumerate(spec.create_simulated_features()):
            ann = GroupAnnotation(
                label=group_counter,
                isotopologue_group=isotopologue_group,
                isotopologue_index=isotopologue_index,
                charge=spec.charge,
            )

            descriptors = {"mz": ft.mz, "rt": ft.rt}
            group = FeatureGroup(group=group_counter, annotation=ann, descriptors=descriptors)
            feature_groups.append(group)
            group_counter += 1
    return feature_groups
